# -*- coding: utf-8 -*-
"""
mimetype interface module.
"""

from abc import abstractmethod
from threading import Lock

from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.core.structs import MultiSingletonMeta, CoreObject


class MIMETypeHandlerSingletonMeta(MultiSingletonMeta):
    """
    mimetype handler singleton meta class.

    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class AbstractMIMETypeHandlerBase(CoreObject, metaclass=MIMETypeHandlerSingletonMeta):
    """
    abstract mimetype handler base class.
    """

    @abstractmethod
    def mimetype(self, value, **options):
        """
        gets the mimetype of given value.

        returns None if it fails to detect the mimetype.

        :param object value: value to detect its mimetype.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def set_next(self, mimetype_handler):
        """
        sets the next mimetype handler and returns it.

        :param AbstractMIMETypeHandlerBase mimetype_handler: mimetype handler instance
                                                             to be set as next handler.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: AbstractMIMETypeHandlerBase
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def accepted_type(self):
        """
        gets the accepted type for this mimetype handler.

        which could detect mimetype from this type.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: type
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
serializer decorators module.
"""

import pyrin.converters.serializer.services as serializer_services


def serializer(*args, **kwargs):
    """
    decorator to register a serializer.

    :param object args: serializer class constructor arguments.
    :param object kwargs: serializer class constructor keyword arguments.

    :keyword bool replace: specifies that if there is another registered
                           serializer with the same accepted type,
                           replace it with the new one, otherwise raise
                           an error. defaults to False.

    :raises InvalidSerializerTypeError: invalid serializer type error.
    :raises DuplicatedSerializerError: duplicated serializer error.

    :returns: serializer class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available serializers.

        :param type cls: serializer class.

        :returns: serializer class.
        :rtype: type
        """

        instance = cls(*args, **kwargs)
        serializer_services.register_serializer(instance, **kwargs)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
orm query base module.
"""

import inspect

from sqlalchemy.orm import Query, lazyload
from sqlalchemy import inspection, log, func, literal, distinct, select
from sqlalchemy.orm.attributes import InstrumentedAttribute

import pyrin.utils.misc as misc_utils
import pyrin.utils.sqlalchemy as sqlalchemy_utils
import pyrin.database.paging.services as paging_services
import pyrin.security.session.services as session_services

from pyrin.core.globals import _, SECURE_FALSE, SECURE_TRUE
from pyrin.database.model.base import BaseEntity
from pyrin.database.orm.sql.schema.base import CoreColumn
from pyrin.database.services import get_current_store
from pyrin.database.orm.query.exceptions import ColumnsOutOfScopeError, \
    InvalidOrderByScopeError, EfficientCountIsNotPossibleError


@inspection._self_inspects
@log.class_logger
class CoreQuery(Query):
    """
    core query class.

    this is the application default query class.
    it extends sqlalchemy `Query` class.
    """

    def __init__(self, entities, session=None, **options):
        """
        initializes an instance of CoreQuery.

        this method has been overridden to provide the concept of scope to queries.
        it is useful if you want to let users (end users not developers) to select
        which columns they want to be returned in a service. in this situation, if
        there is no scope defined, they could add any columns of other entities, but
        using scope, prevents this. but on normal use cases there is no need to define
        scope, and its also more efficient.

        :param tuple entities: entities or columns that are needed for query.
        :param Session session: optional session object to bind this query to it.

        :keyword type | tuple[type] scope: class type of the entities that this
                                           query instance will work on. if the
                                           query is working on multiple entities,
                                           this value must be a tuple of all class
                                           types of that entities.

                                           for example: if you set
                                           `entities=SomeEntity.id, AnotherEntity.name`
                                           you should leave `scope=None` to skip validation
                                           or you could set
                                           `scope=(SomeEntity, AnotherEntity)`
                                           this way validation succeeds, but if
                                           you set `scope=SomeEntity`
                                           then the query will not be executed
                                           and an error will be raised.

        :raises ColumnsOutOfScopeError: columns out of scope error.
        """

        scope = options.get('scope', None)
        if scope is not None:
            self._prepare_validation(entities, scope)

        super().__init__(entities, session)

    def _validate_scope(self, entities, scope):
        """
        validates the query for given entities based on provided scope.

        :param tuple entities: entities or columns that needed for query.

        :param type[BaseEntity] | tuple[type[BaseEntity]] scope: class type of the entities
                                                                 that this query instance will
                                                                 work on. if the query is working
                                                                 on multiple entities, this value
                                                                 must be a tuple of all class
                                                                 types of that entities.

        :raises ColumnsOutOfScopeError: columns out of scope error.
        """

        scope_classes = set(entity for entity in scope if inspect.isclass(entity)
                            and issubclass(entity, BaseEntity))

        requested_classes = set(entity for entity in entities if inspect.isclass(entity)
                                and issubclass(entity, BaseEntity))

        requested_classes_by_column = set(column.class_ for column in entities if
                                          isinstance(column, InstrumentedAttribute)
                                          and issubclass(getattr(column, 'class_', type),
                                                         BaseEntity))

        all_requested_classes = requested_classes.union(requested_classes_by_column)

        if not all_requested_classes.issubset(scope_classes):
            raise ColumnsOutOfScopeError(_('Requested columns are out of scope of this '
                                           'query. Please revise the requested columns.'))

    def _prepare_validation(self, entities, scope):
        """
        prepares entities and scope and validates them if required.

        :param tuple entities: entities or columns that needed for query.

        :param type[BaseEntity] | tuple[type[BaseEntity]] scope: class type of the entities that
                                                                 this query instance will work on.
                                                                 if the query is working on
                                                                 multiple entities, this value
                                                                 must be a tuple of all class
                                                                 types of that entities.

        :raises ColumnsOutOfScopeError: columns out of scope error.
        """

        if scope is None or entities is None:
            return

        scope = misc_utils.make_iterable(scope, tuple)
        entities = misc_utils.make_iterable(entities, tuple)

        if len(scope) <= 0 or len(entities) <= 0:
            return

        self._validate_scope(entities, scope)

    def _copy_statement(self, old_statement, *new_columns):
        """
        gets a copy of given statement with columns replaced with new columns.

        this method is used to generate statement for `_count` function.

        :param Select old_statement: the original statement to make a copy of it.
        :param CoreColumn | str new_columns: new columns to be used in copied statement.

        :rtype: Select
        """

        statement = select(*new_columns).select_from(*old_statement.froms)
        statement._where_criteria = old_statement._where_criteria
        statement._group_by_clauses = old_statement._group_by_clauses
        statement._having_criteria = old_statement._having_criteria
        statement._distinct_on = old_statement._distinct_on
        statement._distinct = old_statement._distinct
        statement._limit_clause = old_statement._limit_clause
        statement._offset_clause = old_statement._offset_clause
        statement._order_by_clauses = old_statement._order_by_clauses
        statement._correlate = old_statement._correlate
        statement._correlate_except = old_statement._correlate_except
        statement._auto_correlate = old_statement._auto_correlate

        return statement

    def _get_appropriate_column(self, *columns):
        """
        gets the appropriate column to be used in `_count` function.

        it may return None if no appropriate column could be found.

        :param CoreColumn | object columns: columns to choose from.

        :rtype: CoreColumn | str
        """

        if columns is None or len(columns) <= 0 or columns[0] is None:
            return None

        if len(columns) == 1:
            return columns[0]

        for item in columns:
            if not isinstance(item, CoreColumn):
                return None

        return '*'

    def _count(self, **options):
        """
        returns the count of rows that the sql formed by this `Query` would return.

        this method generates a single sql query like below:
        select count(column, ...)
        from table
        where ...

        if a single query could not be produced, an error may be raised.

        :keyword CoreColumn column: column to be used in count function.
                                    defaults to `*` if not provided.

        :keyword bool distinct: specifies that count should
                                be executed on distinct select.
                                defaults to False if not provided.
                                note that `distinct` will only be
                                used if `column` is also provided.

        :raises EfficientCountIsNotPossibleError: efficient count is not possible error.

        :rtype: int
        """

        old_statement = self.options(lazyload('*')).statement
        if not old_statement.is_select or not old_statement.is_selectable:
            raise EfficientCountIsNotPossibleError('The provided statement is not a select '
                                                   'statement and efficient count could not '
                                                   'be produced for it.')

        if old_statement._group_by_clauses is not None \
                and len(old_statement._group_by_clauses) > 0:
            raise EfficientCountIsNotPossibleError('The provided statement has group by clause '
                                                   'and efficient count could not be produced '
                                                   'for it.')

        column = options.get('column')
        is_distinct = options.get('distinct', False)
        func_count = None
        if column is None:
            column = self._get_appropriate_column(*old_statement.selected_columns)
            if column is None:
                raise EfficientCountIsNotPossibleError('The provided statement has complex '
                                                       'columns and efficient count could not '
                                                       'be produced for it.')
            func_count = func.count(column)
        else:
            if is_distinct is True:
                func_count = func.count(distinct(column))
            else:
                func_count = func.count(column)

        statement = self._copy_statement(old_statement, func_count).order_by(None)
        store = get_current_store()
        result = store.execute(statement).scalar()
        if result is None:
            result = 0

        return result

    def count(self, **options):
        """
        returns the count of rows that the sql formed by this `Query` would return.

        this method is overridden to prevent inefficient count() of sqlalchemy `Query`
        which produces a subquery, if possible. otherwise it uses the default sqlalchemy
        `count` method.

        :keyword CoreColumn column: column to be used in count function.
                                    defaults to `*` if not provided.
                                    this is only used if a single query
                                    could be produced for count.

        :keyword bool distinct: specifies that count should
                                be executed on distinct select.
                                defaults to False if not provided.
                                note that `distinct` will only be
                                used if `column` is also provided.

        :rtype: int
        """

        try:
            return self._count(**options)
        except Exception:
            return super().count()

    def paginate(self, **options):
        """
        sets the offset and limit for current query.

        the offset and limit values will be extracted from given inputs.
        note that `.paginate` must be called after all other query methods
        have been called. otherwise unexpected behaviour may occur.

        if there is no request context available or there is no paginator for
        current request, pagination will be ignored. this is required to prevent
        security risks by clients.

        if you want pagination outside of request context, you could use
        `.limit` and `.offset` methods directly.

        :keyword SECURE_TRUE | SECURE_FALSE inject_total: inject total count into
                                                          current request.
                                                          defaults to `SECURE_FALSE`
                                                          if not provided.

        :keyword CoreColumn column: column to be used in count function.
                                    defaults to `*` if not provided.
                                    this is only used if `inject_total` is
                                    provided and a single query could be
                                    produced for count.

        :keyword bool distinct: specifies that count should
                                be executed on distinct select.
                                defaults to False if not provided.
                                note that `distinct` will only be
                                used if `column` is also provided.

        :keyword int __limit__: limit value.
        :keyword int __offset__: offset value.
        """

        inject_total = options.get('inject_total', SECURE_FALSE)
        paginator = None
        if session_services.is_request_context_available() is True:
            paginator = session_services.get_request_context('paginator', None)
        else:
            return self

        if paginator is not None:
            if inject_total is SECURE_TRUE:
                paginator.total_count = self.order_by(None).count(**options)

            limit, offset = paging_services.get_paging_keys(**options)
            return self.limit(limit).offset(offset)

        return self

    def existed(self):
        """
        gets a value indicating that current query has any results.

        this is a helper method to simplify the use of `.exists` method of sqlalchemy.

        :rtype: bool
        """

        store = get_current_store()
        result = store.query(literal(True)).filter(self.exists()).scalar()
        return result is True

    def delete(self, synchronize_session=False):
        """
        performs a bulk delete query.

        this method is overridden to provide the most performant
        `synchronize_session` value as default value.

        :param str | bool synchronize_session: session synchronization strategy.
                                               it could be set to False, `fetch`
                                               or `evaluate`. defaults to False.

        :returns: count of affected rows
        :rtype: int
        """

        return super().delete(synchronize_session=synchronize_session)

    def update(self, values, synchronize_session=False, update_args=None):
        """
        performs a bulk update query.

        this method is overridden to provide the most performant
        `synchronize_session` value as default value.

        :param values: a dictionary with attributes names, or alternatively
                       mapped attributes or sql expressions, as keys, and
                       literal values or sql expressions as values.

        :param str | bool synchronize_session: session synchronization strategy.
                                               it could be set to False, `fetch`
                                               or `evaluate`. defaults to False.

        :param update_args: optional dictionary, if present will be passed
                            to the underlying `_expression.update` construct
                            as the `**kw` for the object.

        :returns: count of affected rows
        :rtype: int
        """

        return super().update(values,
                              synchronize_session=synchronize_session,
                              update_args=update_args)

    def safe_order_by(self, scope, *force_order, **options):
        """
        apply one or more `ORDER BY` criterion to the query and return new `CoreQuery`.

        this method only accepts column names, not instances.
        it is implemented to be used for ordering by client inputs.

        this method will automatically fetch order by columns from `order_by`
        keyword if it is provided in options.

        default ordering is ascending, but it could be changed to descending
        by prefixing `-` to column names.

        for example:

        name, +age -> ordering for name and age columns both ascending.
        name, -age -> ordering for name ascending and age descending.

        if column names are not valid, this method ignores them and does not produce
        invalid order by expression. if you do not want to ignore invalid columns,
        use `order_by` method instead.

        :param type[BaseEntity] | list[str] scope: entity class or a list of column names
                                                   to pick order by columns from it.
                                                   note that if a list is provided, column
                                                   names must be the name of actual table
                                                   columns.

        :param str force_order: column names to be appended to `order_by` columns.
                                note that they must be the attribute names
                                of entity if scope is an entity class, otherwise
                                they must be the actual table column names.

        :keyword list[str] | str order_by: column names to be used in order by criterion.
                                           this value is defined to let clients directly
                                           provide order by columns to services through
                                           options. if `order_by` is provided, and
                                           `force_order` is also provided, `force_order`
                                           will be appended to `order_by` values.
                                           it is useful if you want to assure that always
                                           a valid order by will be generated even if
                                           client does not provide any column names.
                                           note that column names must be the attribute names
                                           of entity if scope is an entity class, otherwise
                                           they must be the actual table column names.

        :raises InvalidOrderByScopeError: invalid order by scope error.

        :rtype: CoreQuery
        """

        criterion = None
        columns = options.get('order_by')
        columns = misc_utils.make_iterable(columns, list)
        columns.extend(force_order)

        if isinstance(scope, type) and issubclass(scope, BaseEntity):
            criterion = scope.get_ordering_criterion(*columns, ignore_invalid=True)

        elif isinstance(scope, list):
            criterion = sqlalchemy_utils.get_ordering_criterion(*columns,
                                                                valid_columns=scope,
                                                                ignore_invalid=True)
        else:
            raise InvalidOrderByScopeError('Order by "scope" must be an entity '
                                           'class or a list of column names.')

        return self.order_by(*criterion)



# -*- coding: utf-8 -*-
"""
caching local package.
"""


from pyrin.packaging.base import Package


class CachingLocalPackage(Package):
    """
    caching local package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
deserializer handlers dictionary module.
"""

import re

import flask.json as flask_json

import pyrin.converters.deserializer.services as deserializer_services

from pyrin.converters.deserializer.decorators import deserializer
from pyrin.core.structs import DTO
from pyrin.core.globals import NULL
from pyrin.converters.deserializer.handlers.base import DeserializerBase, \
    StringPatternDeserializerBase


@deserializer()
class DictionaryDeserializer(DeserializerBase):
    """
    dictionary deserializer class.
    """

    def __init__(self, **options):
        """
        creates an instance of DictionaryDeserializer.

        :keyword bool internal: specifies that this deserializer is internal.
                                internal deserializers will not be used for
                                deserializing client inputs.
                                defaults to False if not provided.
        """

        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes every possible value available in input dictionary.

        gets a new deserialized dictionary, leaving the input unchanged.

        :param dict value: value that should be deserialized.

        :rtype: dict
        """

        result = DTO(**value)
        for key, item in result.items():
            result[key] = deserializer_services.deserialize(item, **options)

        return result

    @property
    def accepted_type(self):
        """
        gets the accepted type for this deserializer.

        which could deserialize values from this type.

        :rtype: type[dict]
        """

        return dict


@deserializer()
class StringDictionaryDeserializer(StringPatternDeserializerBase):
    """
    string dictionary deserializer class.
    """

    # default min for this deserializer is 2 because
    # it should at least have { and } at both ends.
    DEFAULT_MIN = 2

    # matches a dictionary inside string, all of these values will be matched.
    # example: {}, {22: 1}, {'key1': 1, 'key2': 2}
    DICT_REGEX = re.compile(r'^{.*}$')

    def __init__(self, **options):
        """
        creates an instance of StringDictionaryDeserializer.

        :keyword list[tuple[Pattern, int, int]] accepted_formats: a list of custom accepted
                                                                  patterns and their min and
                                                                  max length for dictionary
                                                                  deserialization.

        :note accepted_formats: list[tuple[Pattern format, int min_length, int max_length]]

        :keyword bool internal: specifies that this deserializer is internal.
                                internal deserializers will not be used for
                                deserializing client inputs.
                                defaults to False if not provided.
        """

        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes the given value.

        returns `NULL` object if deserialization fails.

        :param str value: value to be deserialized.

        :keyword Pattern matching_pattern: the pattern that has matched the value.

        :rtype: dict
        """

        try:
            dict_value = flask_json.loads(value)
            if dict_value is not None:
                return deserializer_services.deserialize(dict_value, **options)

            return NULL
        except Exception:
            return NULL

    @property
    def default_formats(self):
        """
        gets default accepted patterns that this deserializer could deserialize value from.

        :returns: list[tuple[Pattern format, int min_length, int max_length]]
        :rtype: list[tuple[Pattern, int, int]]
        """

        return [(self.DICT_REGEX, self.UNDEF_LENGTH, self.UNDEF_LENGTH)]



# -*- coding: utf-8 -*-
"""
json encoder module.
"""

from datetime import datetime, date, time

from flask.json import JSONEncoder

import pyrin.globalization.datetime.services as datetime_services

from pyrin.utils import encoding
from pyrin.database.model.base import BaseEntity


class CoreJSONEncoder(JSONEncoder):
    """
    the default pyrin json encoder.

    it extends the default flask json encoder to get the
    correct string representation for other complex types.
    """

    def default(self, o):
        """
        implement this method in a subclass such that it returns a
        serializable object for `o`, or calls the base implementation (to
        raise a `TypeError`).

        this method is overridden to be able to serialize complex python
        types to json string.

        :returns: serializable object
        """

        if isinstance(o, datetime):
            return datetime_services.to_datetime_string(o, to_server=False, from_server=True)
        if isinstance(o, date):
            return datetime_services.to_date_string(o)
        if isinstance(o, time):
            return datetime_services.to_time_string(o, to_server=False, from_server=True)
        if isinstance(o, bytes):
            return encoding.bytes_to_base64_string(o)
        if isinstance(o, BaseEntity):
            return o.to_dict()

        return super().default(o)



# -*- coding: utf-8 -*-
"""
orm types mixin module.
"""

from datetime import datetime

import pyrin.globalization.datetime.services as datetime_services
import pyrin.utils.datetime as datetime_utils

from pyrin.database.enumerations import DialectEnum
from pyrin.database.orm.types.base import CoreCustomType


class DateTimeMixin(CoreCustomType):
    """
    datetime mixin class.

    this is a helper class that adds ability to handle datetime
    values correctly on sqlite backend.
    on other backends, it works as default.
    """

    def _to_database(self, value, dialect):
        """
        converts given value to be emitted to database.

        :param datetime | date value: value to be processed.
        :param Dialect dialect: the dialect in use.

        :rtype: datetime
        """

        if dialect.name == DialectEnum.SQLITE:
            if not isinstance(value, datetime):
                value = datetime_utils.to_datetime_from_date(value)

            return datetime_services.convert_to_utc(value, from_server=True)

        return value

    def _from_database(self, value, dialect):
        """
        converts given value to python type after fetching it from database.

        :param datetime | date value: value to be processed.
        :param Dialect dialect: the dialect in use.


        :rtype: datetime
        """

        if dialect.name == DialectEnum.SQLITE:
            return datetime_services.convert_from_utc(value, to_server=True)

        return value

    def _coerce_to_string(self, value, dialect):
        """
        coerces the given value to string before sending to database.

        :param datetime | date value: value to be processed.
        :param Dialect dialect: the dialect in use.

        :rtype: str
        """

        if dialect.name == DialectEnum.SQLITE:
            return datetime_services.get_timestamp(value, microsecond=True)

        return value

    @property
    def python_type(self):
        """
        gets the python type object expected to be returned by instances of this type.

        :rtype: type[datetime]
        """

        return datetime



# -*- coding: utf-8 -*-
"""
datetime enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class TimezoneEnum(CoreEnum):
    """
    timezone enum.
    """

    AFRICA_ABIDJAN = 'Africa/Abidjan'
    AFRICA_ACCRA = 'Africa/Accra'
    AFRICA_ADDIS_ABABA = 'Africa/Addis_Ababa'
    AFRICA_ALGIERS = 'Africa/Algiers'
    AFRICA_ASMARA = 'Africa/Asmara'
    AFRICA_ASMERA = 'Africa/Asmera'
    AFRICA_BAMAKO = 'Africa/Bamako'
    AFRICA_BANGUI = 'Africa/Bangui'
    AFRICA_BANJUL = 'Africa/Banjul'
    AFRICA_BISSAU = 'Africa/Bissau'
    AFRICA_BLANTYRE = 'Africa/Blantyre'
    AFRICA_BRAZZAVILLE = 'Africa/Brazzaville'
    AFRICA_BUJUMBURA = 'Africa/Bujumbura'
    AFRICA_CAIRO = 'Africa/Cairo'
    AFRICA_CASABLANCA = 'Africa/Casablanca'
    AFRICA_CEUTA = 'Africa/Ceuta'
    AFRICA_CONAKRY = 'Africa/Conakry'
    AFRICA_DAKAR = 'Africa/Dakar'
    AFRICA_DAR_ES_SALAAM = 'Africa/Dar_es_Salaam'
    AFRICA_DJIBOUTI = 'Africa/Djibouti'
    AFRICA_DOUALA = 'Africa/Douala'
    AFRICA_EL_AAIUN = 'Africa/El_Aaiun'
    AFRICA_FREETOWN = 'Africa/Freetown'
    AFRICA_GABORONE = 'Africa/Gaborone'
    AFRICA_HARARE = 'Africa/Harare'
    AFRICA_JOHANNESBURG = 'Africa/Johannesburg'
    AFRICA_JUBA = 'Africa/Juba'
    AFRICA_KAMPALA = 'Africa/Kampala'
    AFRICA_KHARTOUM = 'Africa/Khartoum'
    AFRICA_KIGALI = 'Africa/Kigali'
    AFRICA_KINSHASA = 'Africa/Kinshasa'
    AFRICA_LAGOS = 'Africa/Lagos'
    AFRICA_LIBREVILLE = 'Africa/Libreville'
    AFRICA_LOME = 'Africa/Lome'
    AFRICA_LUANDA = 'Africa/Luanda'
    AFRICA_LUBUMBASHI = 'Africa/Lubumbashi'
    AFRICA_LUSAKA = 'Africa/Lusaka'
    AFRICA_MALABO = 'Africa/Malabo'
    AFRICA_MAPUTO = 'Africa/Maputo'
    AFRICA_MASERU = 'Africa/Maseru'
    AFRICA_MBABANE = 'Africa/Mbabane'
    AFRICA_MOGADISHU = 'Africa/Mogadishu'
    AFRICA_MONROVIA = 'Africa/Monrovia'
    AFRICA_NAIROBI = 'Africa/Nairobi'
    AFRICA_NDJAMENA = 'Africa/Ndjamena'
    AFRICA_NIAMEY = 'Africa/Niamey'
    AFRICA_NOUAKCHOTT = 'Africa/Nouakchott'
    AFRICA_OUAGADOUGOU = 'Africa/Ouagadougou'
    AFRICA_PORTO_NOVO = 'Africa/Porto-Novo'
    AFRICA_SAO_TOME = 'Africa/Sao_Tome'
    AFRICA_TIMBUKTU = 'Africa/Timbuktu'
    AFRICA_TRIPOLI = 'Africa/Tripoli'
    AFRICA_TUNIS = 'Africa/Tunis'
    AFRICA_WINDHOEK = 'Africa/Windhoek'
    AMERICA_ADAK = 'America/Adak'
    AMERICA_ANCHORAGE = 'America/Anchorage'
    AMERICA_ANGUILLA = 'America/Anguilla'
    AMERICA_ANTIGUA = 'America/Antigua'
    AMERICA_ARAGUAINA = 'America/Araguaina'
    AMERICA_ARGENTINA_BUENOS_AIRES = 'America/Argentina/Buenos_Aires'
    AMERICA_ARGENTINA_CATAMARCA = 'America/Argentina/Catamarca'
    AMERICA_ARGENTINA_COMODRIVADAVIA = 'America/Argentina/ComodRivadavia'
    AMERICA_ARGENTINA_CORDOBA = 'America/Argentina/Cordoba'
    AMERICA_ARGENTINA_JUJUY = 'America/Argentina/Jujuy'
    AMERICA_ARGENTINA_LA_RIOJA = 'America/Argentina/La_Rioja'
    AMERICA_ARGENTINA_MENDOZA = 'America/Argentina/Mendoza'
    AMERICA_ARGENTINA_RIO_GALLEGOS = 'America/Argentina/Rio_Gallegos'
    AMERICA_ARGENTINA_SALTA = 'America/Argentina/Salta'
    AMERICA_ARGENTINA_SAN_JUAN = 'America/Argentina/San_Juan'
    AMERICA_ARGENTINA_SAN_LUIS = 'America/Argentina/San_Luis'
    AMERICA_ARGENTINA_TUCUMAN = 'America/Argentina/Tucuman'
    AMERICA_ARGENTINA_USHUAIA = 'America/Argentina/Ushuaia'
    AMERICA_ARUBA = 'America/Aruba'
    AMERICA_ASUNCION = 'America/Asuncion'
    AMERICA_ATIKOKAN = 'America/Atikokan'
    AMERICA_ATKA = 'America/Atka'
    AMERICA_BAHIA = 'America/Bahia'
    AMERICA_BAHIA_BANDERAS = 'America/Bahia_Banderas'
    AMERICA_BARBADOS = 'America/Barbados'
    AMERICA_BELEM = 'America/Belem'
    AMERICA_BELIZE = 'America/Belize'
    AMERICA_BLANC_SABLON = 'America/Blanc-Sablon'
    AMERICA_BOA_VISTA = 'America/Boa_Vista'
    AMERICA_BOGOTA = 'America/Bogota'
    AMERICA_BOISE = 'America/Boise'
    AMERICA_BUENOS_AIRES = 'America/Buenos_Aires'
    AMERICA_CAMBRIDGE_BAY = 'America/Cambridge_Bay'
    AMERICA_CAMPO_GRANDE = 'America/Campo_Grande'
    AMERICA_CANCUN = 'America/Cancun'
    AMERICA_CARACAS = 'America/Caracas'
    AMERICA_CATAMARCA = 'America/Catamarca'
    AMERICA_CAYENNE = 'America/Cayenne'
    AMERICA_CAYMAN = 'America/Cayman'
    AMERICA_CHICAGO = 'America/Chicago'
    AMERICA_CHIHUAHUA = 'America/Chihuahua'
    AMERICA_CORAL_HARBOUR = 'America/Coral_Harbour'
    AMERICA_CORDOBA = 'America/Cordoba'
    AMERICA_COSTA_RICA = 'America/Costa_Rica'
    AMERICA_CRESTON = 'America/Creston'
    AMERICA_CUIABA = 'America/Cuiaba'
    AMERICA_CURACAO = 'America/Curacao'
    AMERICA_DANMARKSHAVN = 'America/Danmarkshavn'
    AMERICA_DAWSON = 'America/Dawson'
    AMERICA_DAWSON_CREEK = 'America/Dawson_Creek'
    AMERICA_DENVER = 'America/Denver'
    AMERICA_DETROIT = 'America/Detroit'
    AMERICA_DOMINICA = 'America/Dominica'
    AMERICA_EDMONTON = 'America/Edmonton'
    AMERICA_EIRUNEPE = 'America/Eirunepe'
    AMERICA_EL_SALVADOR = 'America/El_Salvador'
    AMERICA_ENSENADA = 'America/Ensenada'
    AMERICA_FORT_NELSON = 'America/Fort_Nelson'
    AMERICA_FORT_WAYNE = 'America/Fort_Wayne'
    AMERICA_FORTALEZA = 'America/Fortaleza'
    AMERICA_GLACE_BAY = 'America/Glace_Bay'
    AMERICA_GODTHAB = 'America/Godthab'
    AMERICA_GOOSE_BAY = 'America/Goose_Bay'
    AMERICA_GRAND_TURK = 'America/Grand_Turk'
    AMERICA_GRENADA = 'America/Grenada'
    AMERICA_GUADELOUPE = 'America/Guadeloupe'
    AMERICA_GUATEMALA = 'America/Guatemala'
    AMERICA_GUAYAQUIL = 'America/Guayaquil'
    AMERICA_GUYANA = 'America/Guyana'
    AMERICA_HALIFAX = 'America/Halifax'
    AMERICA_HAVANA = 'America/Havana'
    AMERICA_HERMOSILLO = 'America/Hermosillo'
    AMERICA_INDIANA_INDIANAPOLIS = 'America/Indiana/Indianapolis'
    AMERICA_INDIANA_KNOX = 'America/Indiana/Knox'
    AMERICA_INDIANA_MARENGO = 'America/Indiana/Marengo'
    AMERICA_INDIANA_PETERSBURG = 'America/Indiana/Petersburg'
    AMERICA_INDIANA_TELL_CITY = 'America/Indiana/Tell_City'
    AMERICA_INDIANA_VEVAY = 'America/Indiana/Vevay'
    AMERICA_INDIANA_VINCENNES = 'America/Indiana/Vincennes'
    AMERICA_INDIANA_WINAMAC = 'America/Indiana/Winamac'
    AMERICA_INDIANAPOLIS = 'America/Indianapolis'
    AMERICA_INUVIK = 'America/Inuvik'
    AMERICA_IQALUIT = 'America/Iqaluit'
    AMERICA_JAMAICA = 'America/Jamaica'
    AMERICA_JUJUY = 'America/Jujuy'
    AMERICA_JUNEAU = 'America/Juneau'
    AMERICA_KENTUCKY_LOUISVILLE = 'America/Kentucky/Louisville'
    AMERICA_KENTUCKY_MONTICELLO = 'America/Kentucky/Monticello'
    AMERICA_KNOX_IN = 'America/Knox_IN'
    AMERICA_KRALENDIJK = 'America/Kralendijk'
    AMERICA_LA_PAZ = 'America/La_Paz'
    AMERICA_LIMA = 'America/Lima'
    AMERICA_LOS_ANGELES = 'America/Los_Angeles'
    AMERICA_LOUISVILLE = 'America/Louisville'
    AMERICA_LOWER_PRINCES = 'America/Lower_Princes'
    AMERICA_MACEIO = 'America/Maceio'
    AMERICA_MANAGUA = 'America/Managua'
    AMERICA_MANAUS = 'America/Manaus'
    AMERICA_MARIGOT = 'America/Marigot'
    AMERICA_MARTINIQUE = 'America/Martinique'
    AMERICA_MATAMOROS = 'America/Matamoros'
    AMERICA_MAZATLAN = 'America/Mazatlan'
    AMERICA_MENDOZA = 'America/Mendoza'
    AMERICA_MENOMINEE = 'America/Menominee'
    AMERICA_MERIDA = 'America/Merida'
    AMERICA_METLAKATLA = 'America/Metlakatla'
    AMERICA_MEXICO_CITY = 'America/Mexico_City'
    AMERICA_MIQUELON = 'America/Miquelon'
    AMERICA_MONCTON = 'America/Moncton'
    AMERICA_MONTERREY = 'America/Monterrey'
    AMERICA_MONTEVIDEO = 'America/Montevideo'
    AMERICA_MONTREAL = 'America/Montreal'
    AMERICA_MONTSERRAT = 'America/Montserrat'
    AMERICA_NASSAU = 'America/Nassau'
    AMERICA_NEW_YORK = 'America/New_York'
    AMERICA_NIPIGON = 'America/Nipigon'
    AMERICA_NOME = 'America/Nome'
    AMERICA_NORONHA = 'America/Noronha'
    AMERICA_NORTH_DAKOTA_BEULAH = 'America/North_Dakota/Beulah'
    AMERICA_NORTH_DAKOTA_CENTER = 'America/North_Dakota/Center'
    AMERICA_NORTH_DAKOTA_NEW_SALEM = 'America/North_Dakota/New_Salem'
    AMERICA_OJINAGA = 'America/Ojinaga'
    AMERICA_PANAMA = 'America/Panama'
    AMERICA_PANGNIRTUNG = 'America/Pangnirtung'
    AMERICA_PARAMARIBO = 'America/Paramaribo'
    AMERICA_PHOENIX = 'America/Phoenix'
    AMERICA_PORT_AU_PRINCE = 'America/Port-au-Prince'
    AMERICA_PORT_OF_SPAIN = 'America/Port_of_Spain'
    AMERICA_PORTO_ACRE = 'America/Porto_Acre'
    AMERICA_PORTO_VELHO = 'America/Porto_Velho'
    AMERICA_PUERTO_RICO = 'America/Puerto_Rico'
    AMERICA_RAINY_RIVER = 'America/Rainy_River'
    AMERICA_RANKIN_INLET = 'America/Rankin_Inlet'
    AMERICA_RECIFE = 'America/Recife'
    AMERICA_REGINA = 'America/Regina'
    AMERICA_RESOLUTE = 'America/Resolute'
    AMERICA_RIO_BRANCO = 'America/Rio_Branco'
    AMERICA_ROSARIO = 'America/Rosario'
    AMERICA_SANTA_ISABEL = 'America/Santa_Isabel'
    AMERICA_SANTAREM = 'America/Santarem'
    AMERICA_SANTIAGO = 'America/Santiago'
    AMERICA_SANTO_DOMINGO = 'America/Santo_Domingo'
    AMERICA_SAO_PAULO = 'America/Sao_Paulo'
    AMERICA_SCORESBYSUND = 'America/Scoresbysund'
    AMERICA_SHIPROCK = 'America/Shiprock'
    AMERICA_SITKA = 'America/Sitka'
    AMERICA_ST_BARTHELEMY = 'America/St_Barthelemy'
    AMERICA_ST_JOHNS = 'America/St_Johns'
    AMERICA_ST_KITTS = 'America/St_Kitts'
    AMERICA_ST_LUCIA = 'America/St_Lucia'
    AMERICA_ST_THOMAS = 'America/St_Thomas'
    AMERICA_ST_VINCENT = 'America/St_Vincent'
    AMERICA_SWIFT_CURRENT = 'America/Swift_Current'
    AMERICA_TEGUCIGALPA = 'America/Tegucigalpa'
    AMERICA_THULE = 'America/Thule'
    AMERICA_THUNDER_BAY = 'America/Thunder_Bay'
    AMERICA_TIJUANA = 'America/Tijuana'
    AMERICA_TORONTO = 'America/Toronto'
    AMERICA_TORTOLA = 'America/Tortola'
    AMERICA_VANCOUVER = 'America/Vancouver'
    AMERICA_VIRGIN = 'America/Virgin'
    AMERICA_WHITEHORSE = 'America/Whitehorse'
    AMERICA_WINNIPEG = 'America/Winnipeg'
    AMERICA_YAKUTAT = 'America/Yakutat'
    AMERICA_YELLOWKNIFE = 'America/Yellowknife'
    ANTARCTICA_CASEY = 'Antarctica/Casey'
    ANTARCTICA_DAVIS = 'Antarctica/Davis'
    ANTARCTICA_DUMONTDURVILLE = 'Antarctica/DumontDUrville'
    ANTARCTICA_MACQUARIE = 'Antarctica/Macquarie'
    ANTARCTICA_MAWSON = 'Antarctica/Mawson'
    ANTARCTICA_MCMURDO = 'Antarctica/McMurdo'
    ANTARCTICA_PALMER = 'Antarctica/Palmer'
    ANTARCTICA_ROTHERA = 'Antarctica/Rothera'
    ANTARCTICA_SOUTH_POLE = 'Antarctica/South_Pole'
    ANTARCTICA_SYOWA = 'Antarctica/Syowa'
    ANTARCTICA_TROLL = 'Antarctica/Troll'
    ANTARCTICA_VOSTOK = 'Antarctica/Vostok'
    ARCTIC_LONGYEARBYEN = 'Arctic/Longyearbyen'
    ASIA_ADEN = 'Asia/Aden'
    ASIA_ALMATY = 'Asia/Almaty'
    ASIA_AMMAN = 'Asia/Amman'
    ASIA_ANADYR = 'Asia/Anadyr'
    ASIA_AQTAU = 'Asia/Aqtau'
    ASIA_AQTOBE = 'Asia/Aqtobe'
    ASIA_ASHGABAT = 'Asia/Ashgabat'
    ASIA_ASHKHABAD = 'Asia/Ashkhabad'
    ASIA_BAGHDAD = 'Asia/Baghdad'
    ASIA_BAHRAIN = 'Asia/Bahrain'
    ASIA_BAKU = 'Asia/Baku'
    ASIA_BANGKOK = 'Asia/Bangkok'
    ASIA_BARNAUL = 'Asia/Barnaul'
    ASIA_BEIRUT = 'Asia/Beirut'
    ASIA_BISHKEK = 'Asia/Bishkek'
    ASIA_BRUNEI = 'Asia/Brunei'
    ASIA_CALCUTTA = 'Asia/Calcutta'
    ASIA_CHITA = 'Asia/Chita'
    ASIA_CHOIBALSAN = 'Asia/Choibalsan'
    ASIA_CHONGQING = 'Asia/Chongqing'
    ASIA_CHUNGKING = 'Asia/Chungking'
    ASIA_COLOMBO = 'Asia/Colombo'
    ASIA_DACCA = 'Asia/Dacca'
    ASIA_DAMASCUS = 'Asia/Damascus'
    ASIA_DHAKA = 'Asia/Dhaka'
    ASIA_DILI = 'Asia/Dili'
    ASIA_DUBAI = 'Asia/Dubai'
    ASIA_DUSHANBE = 'Asia/Dushanbe'
    ASIA_GAZA = 'Asia/Gaza'
    ASIA_HARBIN = 'Asia/Harbin'
    ASIA_HEBRON = 'Asia/Hebron'
    ASIA_HO_CHI_MINH = 'Asia/Ho_Chi_Minh'
    ASIA_HONG_KONG = 'Asia/Hong_Kong'
    ASIA_HOVD = 'Asia/Hovd'
    ASIA_IRKUTSK = 'Asia/Irkutsk'
    ASIA_ISTANBUL = 'Asia/Istanbul'
    ASIA_JAKARTA = 'Asia/Jakarta'
    ASIA_JAYAPURA = 'Asia/Jayapura'
    ASIA_JERUSALEM = 'Asia/Jerusalem'
    ASIA_KABUL = 'Asia/Kabul'
    ASIA_KAMCHATKA = 'Asia/Kamchatka'
    ASIA_KARACHI = 'Asia/Karachi'
    ASIA_KASHGAR = 'Asia/Kashgar'
    ASIA_KATHMANDU = 'Asia/Kathmandu'
    ASIA_KATMANDU = 'Asia/Katmandu'
    ASIA_KHANDYGA = 'Asia/Khandyga'
    ASIA_KOLKATA = 'Asia/Kolkata'
    ASIA_KRASNOYARSK = 'Asia/Krasnoyarsk'
    ASIA_KUALA_LUMPUR = 'Asia/Kuala_Lumpur'
    ASIA_KUCHING = 'Asia/Kuching'
    ASIA_KUWAIT = 'Asia/Kuwait'
    ASIA_MACAO = 'Asia/Macao'
    ASIA_MACAU = 'Asia/Macau'
    ASIA_MAGADAN = 'Asia/Magadan'
    ASIA_MAKASSAR = 'Asia/Makassar'
    ASIA_MANILA = 'Asia/Manila'
    ASIA_MUSCAT = 'Asia/Muscat'
    ASIA_NICOSIA = 'Asia/Nicosia'
    ASIA_NOVOKUZNETSK = 'Asia/Novokuznetsk'
    ASIA_NOVOSIBIRSK = 'Asia/Novosibirsk'
    ASIA_OMSK = 'Asia/Omsk'
    ASIA_ORAL = 'Asia/Oral'
    ASIA_PHNOM_PENH = 'Asia/Phnom_Penh'
    ASIA_PONTIANAK = 'Asia/Pontianak'
    ASIA_PYONGYANG = 'Asia/Pyongyang'
    ASIA_QATAR = 'Asia/Qatar'
    ASIA_QYZYLORDA = 'Asia/Qyzylorda'
    ASIA_RANGOON = 'Asia/Rangoon'
    ASIA_RIYADH = 'Asia/Riyadh'
    ASIA_SAIGON = 'Asia/Saigon'
    ASIA_SAKHALIN = 'Asia/Sakhalin'
    ASIA_SAMARKAND = 'Asia/Samarkand'
    ASIA_SEOUL = 'Asia/Seoul'
    ASIA_SHANGHAI = 'Asia/Shanghai'
    ASIA_SINGAPORE = 'Asia/Singapore'
    ASIA_SREDNEKOLYMSK = 'Asia/Srednekolymsk'
    ASIA_TAIPEI = 'Asia/Taipei'
    ASIA_TASHKENT = 'Asia/Tashkent'
    ASIA_TBILISI = 'Asia/Tbilisi'
    ASIA_TEHRAN = 'Asia/Tehran'
    ASIA_TEL_AVIV = 'Asia/Tel_Aviv'
    ASIA_THIMBU = 'Asia/Thimbu'
    ASIA_THIMPHU = 'Asia/Thimphu'
    ASIA_TOKYO = 'Asia/Tokyo'
    ASIA_TOMSK = 'Asia/Tomsk'
    ASIA_UJUNG_PANDANG = 'Asia/Ujung_Pandang'
    ASIA_ULAANBAATAR = 'Asia/Ulaanbaatar'
    ASIA_ULAN_BATOR = 'Asia/Ulan_Bator'
    ASIA_URUMQI = 'Asia/Urumqi'
    ASIA_UST_NERA = 'Asia/Ust-Nera'
    ASIA_VIENTIANE = 'Asia/Vientiane'
    ASIA_VLADIVOSTOK = 'Asia/Vladivostok'
    ASIA_YAKUTSK = 'Asia/Yakutsk'
    ASIA_YEKATERINBURG = 'Asia/Yekaterinburg'
    ASIA_YEREVAN = 'Asia/Yerevan'
    ATLANTIC_AZORES = 'Atlantic/Azores'
    ATLANTIC_BERMUDA = 'Atlantic/Bermuda'
    ATLANTIC_CANARY = 'Atlantic/Canary'
    ATLANTIC_CAPE_VERDE = 'Atlantic/Cape_Verde'
    ATLANTIC_FAEROE = 'Atlantic/Faeroe'
    ATLANTIC_FAROE = 'Atlantic/Faroe'
    ATLANTIC_JAN_MAYEN = 'Atlantic/Jan_Mayen'
    ATLANTIC_MADEIRA = 'Atlantic/Madeira'
    ATLANTIC_REYKJAVIK = 'Atlantic/Reykjavik'
    ATLANTIC_SOUTH_GEORGIA = 'Atlantic/South_Georgia'
    ATLANTIC_ST_HELENA = 'Atlantic/St_Helena'
    ATLANTIC_STANLEY = 'Atlantic/Stanley'
    AUSTRALIA_ACT = 'Australia/ACT'
    AUSTRALIA_ADELAIDE = 'Australia/Adelaide'
    AUSTRALIA_BRISBANE = 'Australia/Brisbane'
    AUSTRALIA_BROKEN_HILL = 'Australia/Broken_Hill'
    AUSTRALIA_CANBERRA = 'Australia/Canberra'
    AUSTRALIA_CURRIE = 'Australia/Currie'
    AUSTRALIA_DARWIN = 'Australia/Darwin'
    AUSTRALIA_EUCLA = 'Australia/Eucla'
    AUSTRALIA_HOBART = 'Australia/Hobart'
    AUSTRALIA_LHI = 'Australia/LHI'
    AUSTRALIA_LINDEMAN = 'Australia/Lindeman'
    AUSTRALIA_LORD_HOWE = 'Australia/Lord_Howe'
    AUSTRALIA_MELBOURNE = 'Australia/Melbourne'
    AUSTRALIA_NSW = 'Australia/NSW'
    AUSTRALIA_NORTH = 'Australia/North'
    AUSTRALIA_PERTH = 'Australia/Perth'
    AUSTRALIA_QUEENSLAND = 'Australia/Queensland'
    AUSTRALIA_SOUTH = 'Australia/South'
    AUSTRALIA_SYDNEY = 'Australia/Sydney'
    AUSTRALIA_TASMANIA = 'Australia/Tasmania'
    AUSTRALIA_VICTORIA = 'Australia/Victoria'
    AUSTRALIA_WEST = 'Australia/West'
    AUSTRALIA_YANCOWINNA = 'Australia/Yancowinna'
    BRAZIL_ACRE = 'Brazil/Acre'
    BRAZIL_DENORONHA = 'Brazil/DeNoronha'
    BRAZIL_EAST = 'Brazil/East'
    BRAZIL_WEST = 'Brazil/West'
    CET = 'CET'
    CST6CDT = 'CST6CDT'
    CANADA_ATLANTIC = 'Canada/Atlantic'
    CANADA_CENTRAL = 'Canada/Central'
    CANADA_EAST_SASKATCHEWAN = 'Canada/East-Saskatchewan'
    CANADA_EASTERN = 'Canada/Eastern'
    CANADA_MOUNTAIN = 'Canada/Mountain'
    CANADA_NEWFOUNDLAND = 'Canada/Newfoundland'
    CANADA_PACIFIC = 'Canada/Pacific'
    CANADA_SASKATCHEWAN = 'Canada/Saskatchewan'
    CANADA_YUKON = 'Canada/Yukon'
    CHILE_CONTINENTAL = 'Chile/Continental'
    CHILE_EASTERISLAND = 'Chile/EasterIsland'
    CUBA = 'Cuba'
    EET = 'EET'
    EST = 'EST'
    EST5EDT = 'EST5EDT'
    EGYPT = 'Egypt'
    EIRE = 'Eire'
    ETC_GMT = 'Etc/GMT'
    ETC_GMT_PLUS_0 = 'Etc/GMT+0'
    ETC_GMT_PLUS_1 = 'Etc/GMT+1'
    ETC_GMT_PLUS_10 = 'Etc/GMT+10'
    ETC_GMT_PLUS_11 = 'Etc/GMT+11'
    ETC_GMT_PLUS_12 = 'Etc/GMT+12'
    ETC_GMT_PLUS_2 = 'Etc/GMT+2'
    ETC_GMT_PLUS_3 = 'Etc/GMT+3'
    ETC_GMT_PLUS_4 = 'Etc/GMT+4'
    ETC_GMT_PLUS_5 = 'Etc/GMT+5'
    ETC_GMT_PLUS_6 = 'Etc/GMT+6'
    ETC_GMT_PLUS_7 = 'Etc/GMT+7'
    ETC_GMT_PLUS_8 = 'Etc/GMT+8'
    ETC_GMT_PLUS_9 = 'Etc/GMT+9'
    ETC_GMT_MINUS_0 = 'Etc/GMT-0'
    ETC_GMT_MINUS_1 = 'Etc/GMT-1'
    ETC_GMT_MINUS_10 = 'Etc/GMT-10'
    ETC_GMT_MINUS_11 = 'Etc/GMT-11'
    ETC_GMT_MINUS_12 = 'Etc/GMT-12'
    ETC_GMT_MINUS_13 = 'Etc/GMT-13'
    ETC_GMT_MINUS_14 = 'Etc/GMT-14'
    ETC_GMT_MINUS_2 = 'Etc/GMT-2'
    ETC_GMT_MINUS_3 = 'Etc/GMT-3'
    ETC_GMT_MINUS_4 = 'Etc/GMT-4'
    ETC_GMT_MINUS_5 = 'Etc/GMT-5'
    ETC_GMT_MINUS_6 = 'Etc/GMT-6'
    ETC_GMT_MINUS_7 = 'Etc/GMT-7'
    ETC_GMT_MINUS_8 = 'Etc/GMT-8'
    ETC_GMT_MINUS_9 = 'Etc/GMT-9'
    ETC_GMT0 = 'Etc/GMT0'
    ETC_GREENWICH = 'Etc/Greenwich'
    ETC_UCT = 'Etc/UCT'
    ETC_UTC = 'Etc/UTC'
    ETC_UNIVERSAL = 'Etc/Universal'
    ETC_ZULU = 'Etc/Zulu'
    EUROPE_AMSTERDAM = 'Europe/Amsterdam'
    EUROPE_ANDORRA = 'Europe/Andorra'
    EUROPE_ASTRAKHAN = 'Europe/Astrakhan'
    EUROPE_ATHENS = 'Europe/Athens'
    EUROPE_BELFAST = 'Europe/Belfast'
    EUROPE_BELGRADE = 'Europe/Belgrade'
    EUROPE_BERLIN = 'Europe/Berlin'
    EUROPE_BRATISLAVA = 'Europe/Bratislava'
    EUROPE_BRUSSELS = 'Europe/Brussels'
    EUROPE_BUCHAREST = 'Europe/Bucharest'
    EUROPE_BUDAPEST = 'Europe/Budapest'
    EUROPE_BUSINGEN = 'Europe/Busingen'
    EUROPE_CHISINAU = 'Europe/Chisinau'
    EUROPE_COPENHAGEN = 'Europe/Copenhagen'
    EUROPE_DUBLIN = 'Europe/Dublin'
    EUROPE_GIBRALTAR = 'Europe/Gibraltar'
    EUROPE_GUERNSEY = 'Europe/Guernsey'
    EUROPE_HELSINKI = 'Europe/Helsinki'
    EUROPE_ISLE_OF_MAN = 'Europe/Isle_of_Man'
    EUROPE_ISTANBUL = 'Europe/Istanbul'
    EUROPE_JERSEY = 'Europe/Jersey'
    EUROPE_KALININGRAD = 'Europe/Kaliningrad'
    EUROPE_KIEV = 'Europe/Kiev'
    EUROPE_KIROV = 'Europe/Kirov'
    EUROPE_LISBON = 'Europe/Lisbon'
    EUROPE_LJUBLJANA = 'Europe/Ljubljana'
    EUROPE_LONDON = 'Europe/London'
    EUROPE_LUXEMBOURG = 'Europe/Luxembourg'
    EUROPE_MADRID = 'Europe/Madrid'
    EUROPE_MALTA = 'Europe/Malta'
    EUROPE_MARIEHAMN = 'Europe/Mariehamn'
    EUROPE_MINSK = 'Europe/Minsk'
    EUROPE_MONACO = 'Europe/Monaco'
    EUROPE_MOSCOW = 'Europe/Moscow'
    EUROPE_NICOSIA = 'Europe/Nicosia'
    EUROPE_OSLO = 'Europe/Oslo'
    EUROPE_PARIS = 'Europe/Paris'
    EUROPE_PODGORICA = 'Europe/Podgorica'
    EUROPE_PRAGUE = 'Europe/Prague'
    EUROPE_RIGA = 'Europe/Riga'
    EUROPE_ROME = 'Europe/Rome'
    EUROPE_SAMARA = 'Europe/Samara'
    EUROPE_SAN_MARINO = 'Europe/San_Marino'
    EUROPE_SARAJEVO = 'Europe/Sarajevo'
    EUROPE_SIMFEROPOL = 'Europe/Simferopol'
    EUROPE_SKOPJE = 'Europe/Skopje'
    EUROPE_SOFIA = 'Europe/Sofia'
    EUROPE_STOCKHOLM = 'Europe/Stockholm'
    EUROPE_TALLINN = 'Europe/Tallinn'
    EUROPE_TIRANE = 'Europe/Tirane'
    EUROPE_TIRASPOL = 'Europe/Tiraspol'
    EUROPE_ULYANOVSK = 'Europe/Ulyanovsk'
    EUROPE_UZHGOROD = 'Europe/Uzhgorod'
    EUROPE_VADUZ = 'Europe/Vaduz'
    EUROPE_VATICAN = 'Europe/Vatican'
    EUROPE_VIENNA = 'Europe/Vienna'
    EUROPE_VILNIUS = 'Europe/Vilnius'
    EUROPE_VOLGOGRAD = 'Europe/Volgograd'
    EUROPE_WARSAW = 'Europe/Warsaw'
    EUROPE_ZAGREB = 'Europe/Zagreb'
    EUROPE_ZAPOROZHYE = 'Europe/Zaporozhye'
    EUROPE_ZURICH = 'Europe/Zurich'
    GB = 'GB'
    GB_EIRE = 'GB-Eire'
    GMT = 'GMT'
    GMT_PLUS_0 = 'GMT+0'
    GMT_MINUS_0 = 'GMT-0'
    GMT0 = 'GMT0'
    GREENWICH = 'Greenwich'
    HST = 'HST'
    HONGKONG = 'Hongkong'
    ICELAND = 'Iceland'
    INDIAN_ANTANANARIVO = 'Indian/Antananarivo'
    INDIAN_CHAGOS = 'Indian/Chagos'
    INDIAN_CHRISTMAS = 'Indian/Christmas'
    INDIAN_COCOS = 'Indian/Cocos'
    INDIAN_COMORO = 'Indian/Comoro'
    INDIAN_KERGUELEN = 'Indian/Kerguelen'
    INDIAN_MAHE = 'Indian/Mahe'
    INDIAN_MALDIVES = 'Indian/Maldives'
    INDIAN_MAURITIUS = 'Indian/Mauritius'
    INDIAN_MAYOTTE = 'Indian/Mayotte'
    INDIAN_REUNION = 'Indian/Reunion'
    IRAN = 'Iran'
    ISRAEL = 'Israel'
    JAMAICA = 'Jamaica'
    JAPAN = 'Japan'
    KWAJALEIN = 'Kwajalein'
    LIBYA = 'Libya'
    MET = 'MET'
    MST = 'MST'
    MST7MDT = 'MST7MDT'
    MEXICO_BAJANORTE = 'Mexico/BajaNorte'
    MEXICO_BAJASUR = 'Mexico/BajaSur'
    MEXICO_GENERAL = 'Mexico/General'
    NZ = 'NZ'
    NZ_CHAT = 'NZ-CHAT'
    NAVAJO = 'Navajo'
    PRC = 'PRC'
    PST8PDT = 'PST8PDT'
    PACIFIC_APIA = 'Pacific/Apia'
    PACIFIC_AUCKLAND = 'Pacific/Auckland'
    PACIFIC_BOUGAINVILLE = 'Pacific/Bougainville'
    PACIFIC_CHATHAM = 'Pacific/Chatham'
    PACIFIC_CHUUK = 'Pacific/Chuuk'
    PACIFIC_EASTER = 'Pacific/Easter'
    PACIFIC_EFATE = 'Pacific/Efate'
    PACIFIC_ENDERBURY = 'Pacific/Enderbury'
    PACIFIC_FAKAOFO = 'Pacific/Fakaofo'
    PACIFIC_FIJI = 'Pacific/Fiji'
    PACIFIC_FUNAFUTI = 'Pacific/Funafuti'
    PACIFIC_GALAPAGOS = 'Pacific/Galapagos'
    PACIFIC_GAMBIER = 'Pacific/Gambier'
    PACIFIC_GUADALCANAL = 'Pacific/Guadalcanal'
    PACIFIC_GUAM = 'Pacific/Guam'
    PACIFIC_HONOLULU = 'Pacific/Honolulu'
    PACIFIC_JOHNSTON = 'Pacific/Johnston'
    PACIFIC_KIRITIMATI = 'Pacific/Kiritimati'
    PACIFIC_KOSRAE = 'Pacific/Kosrae'
    PACIFIC_KWAJALEIN = 'Pacific/Kwajalein'
    PACIFIC_MAJURO = 'Pacific/Majuro'
    PACIFIC_MARQUESAS = 'Pacific/Marquesas'
    PACIFIC_MIDWAY = 'Pacific/Midway'
    PACIFIC_NAURU = 'Pacific/Nauru'
    PACIFIC_NIUE = 'Pacific/Niue'
    PACIFIC_NORFOLK = 'Pacific/Norfolk'
    PACIFIC_NOUMEA = 'Pacific/Noumea'
    PACIFIC_PAGO_PAGO = 'Pacific/Pago_Pago'
    PACIFIC_PALAU = 'Pacific/Palau'
    PACIFIC_PITCAIRN = 'Pacific/Pitcairn'
    PACIFIC_POHNPEI = 'Pacific/Pohnpei'
    PACIFIC_PONAPE = 'Pacific/Ponape'
    PACIFIC_PORT_MORESBY = 'Pacific/Port_Moresby'
    PACIFIC_RAROTONGA = 'Pacific/Rarotonga'
    PACIFIC_SAIPAN = 'Pacific/Saipan'
    PACIFIC_SAMOA = 'Pacific/Samoa'
    PACIFIC_TAHITI = 'Pacific/Tahiti'
    PACIFIC_TARAWA = 'Pacific/Tarawa'
    PACIFIC_TONGATAPU = 'Pacific/Tongatapu'
    PACIFIC_TRUK = 'Pacific/Truk'
    PACIFIC_WAKE = 'Pacific/Wake'
    PACIFIC_WALLIS = 'Pacific/Wallis'
    PACIFIC_YAP = 'Pacific/Yap'
    POLAND = 'Poland'
    PORTUGAL = 'Portugal'
    ROC = 'ROC'
    ROK = 'ROK'
    SINGAPORE = 'Singapore'
    TURKEY = 'Turkey'
    UCT = 'UCT'
    US_ALASKA = 'US/Alaska'
    US_ALEUTIAN = 'US/Aleutian'
    US_ARIZONA = 'US/Arizona'
    US_CENTRAL = 'US/Central'
    US_EAST_INDIANA = 'US/East-Indiana'
    US_EASTERN = 'US/Eastern'
    US_HAWAII = 'US/Hawaii'
    US_INDIANA_STARKE = 'US/Indiana-Starke'
    US_MICHIGAN = 'US/Michigan'
    US_MOUNTAIN = 'US/Mountain'
    US_PACIFIC = 'US/Pacific'
    US_PACIFIC_NEW = 'US/Pacific-New'
    US_SAMOA = 'US/Samoa'
    UTC = 'UTC'
    UNIVERSAL = 'Universal'
    W_SU = 'W-SU'
    WET = 'WET'
    ZULU = 'Zulu'



# -*- coding: utf-8 -*-
"""
validator exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class ValidatorManagerException(CoreException):
    """
    validator manager exception.
    """
    pass


class ValidatorManagerBusinessException(CoreBusinessException,
                                        ValidatorManagerException):
    """
    validator manager business exception.
    """
    pass


class ValidationError(ValidatorManagerBusinessException):
    """
    validation error.
    """
    pass


class InvalidValidatorTypeError(ValidatorManagerException):
    """
    invalid validator type error.
    """
    pass


class DuplicatedValidatorError(ValidatorManagerException):
    """
    duplicated validator error.
    """
    pass


class ValidatorNotFoundError(ValidatorManagerException):
    """
    validator not found error.
    """
    pass


class InvalidDataForValidationError(ValidationError):
    """
    invalid data for validation error.
    """
    pass


class InvalidEntityForValidationError(ValidationError):
    """
    invalid entity for validation error.
    """
    pass


class ValidatorDomainNotFoundError(ValidatorManagerException):
    """
    validator domain not found error.
    """
    pass



# -*- coding: utf-8 -*-
"""
orm sql operators package.
"""

from pyrin.packaging.base import Package


class ORMSQLOperatorsPackage(Package):
    """
    orm sql operators package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
encryption interface module.
"""

from abc import abstractmethod
from threading import Lock

from pyrin.core.structs import CoreObject, MultiSingletonMeta
from pyrin.core.exceptions import CoreNotImplementedError


class EncrypterSingletonMeta(MultiSingletonMeta):
    """
    encrypter singleton meta class.
    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class AbstractEncrypterBase(CoreObject, metaclass=EncrypterSingletonMeta):
    """
    abstract encrypter base class.
    all application encrypters must subclass this.
    """

    @abstractmethod
    def encrypt(self, text, **options):
        """
        encrypts the given value and returns the full encrypted
        result which includes the handler name.

        :param str text: text to be encrypted.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def decrypt(self, full_encrypted_value, **options):
        """
        decrypts the given full encrypted value and returns the decrypted result.

        :param str full_encrypted_value: full encrypted value to be decrypted.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def generate_key(self, **options):
        """
        generates a valid key for this handler and returns it.

        :keyword int length: the length of generated key in bytes.
                             note that some encryption handlers may
                             not accept custom key length so this
                             value would be ignored on those handlers.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str | tuple[str, str]
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
datetime component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.globalization.datetime import DateTimePackage
from pyrin.globalization.datetime.manager import DateTimeManager


@component(DateTimePackage.COMPONENT_NAME)
class DateTimeComponent(Component, DateTimeManager):
    """
    datetime component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
caching local handlers package.
"""

from pyrin.packaging.base import Package


class CachingLocalHandlersPackage(Package):
    """
    caching local handlers package class.
    """

    NAME = __name__
    DEPENDS = ['pyrin.configuration',
               'pyrin.globalization.datetime',
               'pyrin.logging']



# -*- coding: utf-8 -*-
"""
alembic decorators module.
"""

import pyrin.database.migration.alembic.services as alembic_services


def alembic_cli_handler(**options):
    """
    decorator to register an alembic cli handler.

    :keyword bool replace: specifies that if there is another registered
                           cli handler with the same name, replace it
                           with the new one, otherwise raise
                           an error. defaults to False.

    :raises InvalidCLIHandlerTypeError: invalid cli handler type error.
    :raises DuplicatedCLIHandlerError: duplicated cli handler error.

    :returns: alembic cli handler class.
    :rtype: AlembicCLIHandlerBase
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available alembic cli handlers.

        :param AlembicCLIHandlerBase cls: alembic cli handler class.

        :returns: alembic cli handler class.
        :rtype: AlembicCLIHandlerBase
        """

        instance = cls()
        alembic_services.register_cli_handler(instance, **options)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
mimetype services module.
"""

from pyrin.processor.mimetype import MIMETypePackage
from pyrin.application.services import get_component


def get_mimetype(value, **options):
    """
    gets the mimetype of given value.

    returns None if it could not detect the correct mimetype.

    :param object value: value to detect its mimetype.

    :returns: mimetype name
    :rtype: str
    """

    return get_component(MIMETypePackage.COMPONENT_NAME).get_mimetype(value, **options)


def register_mimetype_handler(instance, **options):
    """
    registers a new mimetype handler or replaces the existing one.

    if `replace=True` is provided. otherwise, it raises an error
    on adding a mimetype handler which is already registered.

    :param AbstractMIMETypeHandlerBase instance: mimetype handler to be registered.
                                                 it must be an instance of
                                                 AbstractMIMETypeHandlerBase.

    :keyword bool replace: specifies that if there is another registered
                           mimetype handler with the same name and accepted type,
                           replace it with the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidMIMETypeHandlerTypeError: invalid mimetype handler type error.
    :raises DuplicatedMIMETypeHandlerError: duplicated mimetype handler error.
    """

    return get_component(MIMETypePackage.COMPONENT_NAME).register_mimetype_handler(instance,
                                                                                   **options)


def get_mimetype_handlers(**options):
    """
    gets all registered mimetype handlers.

    it could filter mimetype handlers for a specific type if provided.
    it only returns the first mimetype handlers for each type, because
    all mimetype handlers for a given type, are chained together.

    :keyword type accepted_type: specifies to get mimetype handlers which are
                                 registered for the accepted type. if not provided,
                                 all mimetype handlers will be returned.

    :rtype: list[AbstractMIMETypeHandlerBase]
    """

    return get_component(MIMETypePackage.COMPONENT_NAME).get_mimetype_handlers(**options)



# -*- coding: utf-8 -*-
"""
application base module.
"""

import os

from collections import OrderedDict

from time import time

import dotenv

from flask.app import setupmethod
from flask.ctx import has_request_context
from flask import Flask, request as flask_request, _request_ctx_stack as request_stack

import pyrin
import pyrin.converters.serializer.services as serializer_services
import pyrin.packaging.services as packaging_services
import pyrin.configuration.services as config_services
import pyrin.security.authentication.services as authentication_services
import pyrin.security.session.services as session_services
import pyrin.logging.services as logging_services
import pyrin.processor.mimetype.services as mimetype_services
import pyrin.processor.response.services as response_services
import pyrin.utils.misc as misc_utils
import pyrin.utils.path as path_utils
import pyrin.utils.function as function_utils

from pyrin.api.router.structs import CoreURLMap
from pyrin.application.container import _set_app
from pyrin.api.router.handlers.protected import ProtectedRoute
from pyrin.application.enumerations import ApplicationStatusEnum
from pyrin.application.hooks import ApplicationHookBase, PackagingHook
from pyrin.application.mixin import SignalMixin
from pyrin.converters.json.decoder import CoreJSONDecoder
from pyrin.converters.json.encoder import CoreJSONEncoder
from pyrin.core.enumerations import HTTPMethodEnum
from pyrin.core.structs import DTO, Manager, CoreHeaders
from pyrin.core.mixin import HookMixin
from pyrin.database.transaction.contexts import atomic_context
from pyrin.packaging import PackagingPackage
from pyrin.packaging.component import PackagingComponent
from pyrin.settings.static import DEFAULT_COMPONENT_KEY
from pyrin.utils.custom_print import print_warning
from pyrin.utils.dictionary import make_key_upper
from pyrin.processor.mimetype.enumerations import MIMETypeEnum
from pyrin.processor.response.wrappers.base import CoreResponse
from pyrin.processor.request.wrappers.base import CoreRequest
from pyrin.application.structs import ApplicationContext, ApplicationComponent, \
    ApplicationSingletonMeta, Component
from pyrin.application.exceptions import DuplicateContextKeyError, InvalidComponentTypeError, \
    InvalidComponentIDError, DuplicateComponentIDError, DuplicateRouteURLError, \
    InvalidRouteFactoryTypeError, InvalidApplicationStatusError, \
    ApplicationInScriptingModeError, ComponentAttributeError, \
    ApplicationIsNotSubclassedError, InvalidApplicationHookTypeError, \
    OverwritingEndpointIsNotAllowedError, InvalidStaticHostAndHostMatchingError


class Application(Flask, HookMixin, SignalMixin,
                  metaclass=ApplicationSingletonMeta):
    """
    application class.

    server must initialize an instance of a subclass of this class at startup.
    """

    # the application looks for these stores to configure itself.
    # they should be present in settings folder of the upper
    # level application.
    CONFIG_STORES = ['application',
                     'communication',
                     'environment']

    # pyrin default settings path will be registered in application context with this key.
    PYRIN_DEFAULT_SETTINGS_CONTEXT_KEY = 'default_settings_path'

    # settings path will be registered in application context with this key.
    SETTINGS_CONTEXT_KEY = 'settings_path'

    # migrations path will be registered in application context with this key.
    MIGRATIONS_CONTEXT_KEY = 'migrations_path'

    # locale path will be registered in application context with this key.
    LOCALE_CONTEXT_KEY = 'locale_path'

    # pyrin main package path will be registered in application context with this key.
    PYRIN_PATH_CONTEXT_KEY = 'pyrin_path'

    # application main package path will be registered in application context with this key.
    APPLICATION_PATH_CONTEXT_KEY = 'application_path'

    # pyrin root path will be registered in application context with this key.
    # pyrin root path is where pyrin main package is located.
    ROOT_PYRIN_PATH_CONTEXT_KEY = 'root_pyrin_path'

    # application root path will be registered in application context with this key.
    # root path is where application main package and other files are located.
    ROOT_APPLICATION_PATH_CONTEXT_KEY = 'root_application_path'

    # default packaging component to be used by application.
    # if you want to change the default one, you could subclass
    # Application and set `packaging_component_class` to your desired one.
    # this design is compatible with other flask subclassing features.
    # note that your custom packaging component class should not use `@component`
    # decorator to register itself, application will register it instead.
    packaging_component_class = PackagingComponent

    # a function to be used as default response converter when
    # the mimetype of response could not be detected from its body or
    # the founded mimetype could not be handled by pyrin itself.
    # this function should accept a single parameter as response body
    # and extra optional keyword arguments. the founded mimetype will be
    # passed to this function as 'mimetype' keyword argument. metadata for
    # pagination is also passed to this function using 'metadata' keyword.
    # paginator instance will also passed to this function using 'paginator'
    # keyword. the function must return a response object or an object which
    # could be converted to response by flask.
    default_response_converter = None

    headers_class = CoreHeaders
    url_rule_class = ProtectedRoute
    url_map_class = CoreURLMap
    response_class = CoreResponse
    request_class = CoreRequest
    json_decoder = CoreJSONDecoder
    json_encoder = CoreJSONEncoder
    hook_type = ApplicationHookBase
    invalid_hook_type_error = InvalidApplicationHookTypeError

    def __init__(self, **options):
        """
        initializes an instance of Application.

        :keyword str import_name: the name of the application package.
                                  it will be assumed equal to first part
                                  of application package name if not provided.
                                  for example: `pyrin`.
                                  but if the main package name of application includes
                                  more that one package you should provide it manually.
                                  for example: `tests.unit`.

        :keyword str static_url_path: can be used to specify a different path for the
                                      static files on the web. defaults to the name
                                      of the `static_folder` directory.

        :keyword str static_folder: the folder with static files that is served at
                                    `static_url_path`. relative to the application
                                    `root_path` or an absolute path. defaults to `static'.

        :keyword str static_host: the host to use when adding the static route.
                                  defaults to None. required when using `host_matching=True`
                                  with a `static_folder` configured.

        :keyword bool host_matching: set `url_map.host_matching` attribute.
                                     defaults to False.

        :keyword str subdomain_matching: consider the subdomain relative to
                                         `SERVER_NAME` when matching routes.
                                         defaults to False.

        :keyword str template_folder: the folder that contains the templates that should
                                      be used by the application. defaults to
                                      `templates` folder in the root path of the application.

        :keyword str instance_path: an alternative instance path for the application.
                                    by default the folder `instance` next to the
                                    package or module is assumed to be the instance path.

        :keyword bool instance_relative_config: if set to `True` relative filenames
                                                for loading the config are assumed to
                                                be relative to the instance path instead
                                                of the application root.

        :keyword str root_path: flask by default will automatically calculate the path
                                to the root of the application. in certain situations
                                this cannot be achieved (for instance if the package
                                is a python 3 namespace package) and needs to be
                                manually defined.

        :keyword bool scripting_mode: specifies that the application has been started in
                                      scripting mode. application will not be run in
                                      scripting mode and some application hooks will not
                                      get fired when the application has initialized in
                                      scripting mode. defaults to False, if not provided.

        :keyword bool force_json_response: specifies that if response object is
                                           not an html string, consider it as json
                                           and convert it to be json serializable.
                                           even if it is not a dict.
                                           defaults to True if not provided.

        :keyword str settings_directory: settings directory name.
                                         if not provided, defaults to `settings`.

        :keyword str migrations_directory: migrations directory name.
                                           if not provided, defaults to `migrations`.

        :keyword str locale_directory: locale directory name.
                                       if not provided, defaults to `locale`.

        :raises ApplicationIsNotSubclassedError: application is not subclassed error.
        """

        self._assert_is_subclassed()

        self.__status = ApplicationStatusEnum.INITIALIZING
        self._scripting_mode = options.get('scripting_mode', False)
        self._force_json_response = options.get('force_json_response', True)
        self._version = 'Not Provided'

        self._import_name = options.get('import_name', None)
        if self._import_name is not None and (self._import_name == '' or
                                              self._import_name.isspace()):
            self._import_name = None

        flask_kw = self._remove_flask_unrecognized_keywords(**options)
        # we should pass `static_folder=None` to prevent flask from
        # adding static route on startup, then we register it after application is loaded.
        flask_kw.update(static_folder=None)
        super().__init__(self.get_application_name(), **flask_kw)

        # Flask does not call 'super()' in its '__init__()' method.
        # so we have to start initialization of other parents manually.
        HookMixin.__init__(self, **options)

        self._context = ApplicationContext()
        self._components = ApplicationComponent()

        # we have to register some components manually because they are
        # referenced in `application.base` module and could not be loaded
        # automatically because packaging package could not handle them.
        self._register_required_components()

        # setting the application instance in application container module.
        _set_app(self)

        self._register_required_hooks()

        # we should load application at this stage to be able to perform pytest
        # tests after application has been fully loaded. because if we call
        # application.run(), we could not continue execution of other codes.
        self._load(**options)

        self.static_folder = options.get('static_folder', 'static')
        self._register_static_route(options.get('static_host'), self.url_map.host_matching)

    @setupmethod
    def _register_static_route(self, static_host, host_matching, **options):
        """
        registers static route if required.

        :param str static_host: the host to use when adding the static route.
                                defaults to None. required when using `host_matching=True`
                                with a `static_folder` configured.

        :param bool host_matching: set `url_map.host_matching` attribute.
                                   defaults to False.

        :raises InvalidStaticHostAndHostMatchingError: invalid static host and
                                                       host matching error.
        """

        if self.has_static_folder:
            if bool(static_host) != host_matching:
                raise InvalidStaticHostAndHostMatchingError('Invalid static_host/'
                                                            'host_matching combination.')

            self.add_url_rule(self.static_url_path + '/<path:filename>',
                              view_func=self.send_static_file,
                              authenticated=False,
                              endpoint='static',
                              host=static_host)

    def _remove_flask_unrecognized_keywords(self, **options):
        """
        removes any keyword argument which is not recognized by `Flask.__init__()` method.

        it returns a new dict with all recognized keyword arguments.

        :rtype: dict
        """

        result = dict(**options)
        result.pop('import_name', None)
        result.pop('scripting_mode', None)
        result.pop('settings_directory', None)
        result.pop('migrations_directory', None)
        result.pop('locale_directory', None)
        result.pop('force_json_response', None)

        return result

    def _assert_is_subclassed(self):
        """
        asserts that current application instance is subclassed from `Application`.

        and is not a direct instance of `Application` itself.

        :raises ApplicationIsNotSubclassedError: application is not subclassed error.
        """

        if type(self) is Application:
            raise ApplicationIsNotSubclassedError('Current application instance is a direct '
                                                  'instance of [{base}]. you must subclass '
                                                  'from [{base}] in your project and create '
                                                  'an instance of that class to run your '
                                                  'application. this is needed for pyrin to be '
                                                  'able to resolve different paths correctly.'
                                                  .format(base=self))

    def is_scripting_mode(self):
        """
        gets a value indicating that application has been started in scripting mode.

        some application hooks will not fire in this mode. like 'before_application_run'.
        """

        return self._scripting_mode

    @setupmethod
    def _register_required_components(self):
        """
        registers required components that the Application needs
        to reference to them immediately.

        this type of components could not be registered using @component decorator,
        because they will be referenced before Application instance gets initialized.

        note that implementation-wise, application package should depend on other
        packages services as few as possible, so be careful if you needed some external
        packages services inside application package, probably it would be better to
        change your design that enforces the application to use other packages services.
        """

        self.register_component(self.packaging_component_class(
            PackagingPackage.COMPONENT_NAME))

    @setupmethod
    def _register_required_hooks(self):
        """
        registers the required hooks that the Application needs them.

        this type of hooks could not be registered using @hook decorators,
        because they will be referenced before Application instance gets
        initialized.
        """

        packaging_services.register_hook(PackagingHook())

    def _set_status(self, status):
        """
        sets the application status.

        status must be from ApplicationStatusEnum.

        :param str status: application status.

        :enum status:
            INITIALIZING = 'Initializing'
            LOADING = 'Loading'
            READY = 'Ready'
            RUNNING = 'Running'
            TERMINATED = 'Terminated'

        raises InvalidApplicationStatusError: invalid application status error.
        """

        if status not in ApplicationStatusEnum:
            raise InvalidApplicationStatusError('Application status [{state}] is not valid.'
                                                .format(state=status))

        old_status = self.__status
        self.__status = status

        if old_status != status:
            self._application_status_changed(old_status, status)

    def get_status(self):
        """
        gets the application status.

        :rtype: str
        """

        return self.__status

    def add_context(self, key, value, **options):
        """
        adds the given key and it's value into the application context.

        :param str key: related key for storing application context.
        :param object value: related value for storing in application context.

        :keyword bool replace: specifies that if there is already a value with
                               the same key in application context, it should be updated
                               with new value, otherwise raise an error. defaults to False.

        :raises DuplicateContextKeyError: duplicate context key error.
        """

        replace = options.get('replace', False)
        if replace is not True and key in self._context:
            raise DuplicateContextKeyError('Key [{key}] is already available in application '
                                           'context and "replace=True" option is not set, so '
                                           'the new value could not be added.'
                                           .format(key=key))

        self._context[key] = value

    def get_context(self, key, **options):
        """
        gets the application context value that belongs to given key.

        :param str key: key for requested application context.

        :keyword object default: default value to be returned if key is not available.
                                 otherwise, it raises an error if key is not available.

        :raises ContextAttributeError: context attribute error.

        :returns: related value to given key.
        """

        if 'default' not in options:
            return self._context[key]

        default = options.get('default')
        return self._context.get(key, default)

    @setupmethod
    def register_component(self, component, **options):
        """
        registers given application component or replaces the existing one.

        if `replace=True` is provided. otherwise, it raises an error
        on adding an instance which it's id is already available
        in registered components.

        :param Component component: component instance.

        :keyword bool replace: specifies that if there is another registered
                               component with the same id, replace it with the new one.
                               otherwise raise an error. defaults to False.

        :raises InvalidComponentTypeError: invalid component type error.
        :raises InvalidComponentIDError: invalid component id error.
        :raises DuplicateComponentIDError: duplicate component id error.
        """

        if not isinstance(component, Manager):
            raise InvalidComponentTypeError('Input parameter [{component}] is not '
                                            'an instance of [{instance}]. each component '
                                            'class must be subclassed from its respective '
                                            'manager class of the same package and that '
                                            'manager class must be subclassed from [{instance}].'
                                            .format(component=component,
                                                    instance=Manager))

        if not isinstance(component, Component):
            raise InvalidComponentTypeError('Input parameter [{component}] is not '
                                            'an instance of [{instance}].'
                                            .format(component=component,
                                                    instance=Component))

        if not isinstance(component.get_id(), tuple) or \
                len(component.get_id()[0].strip()) == 0:
            raise InvalidComponentIDError('Component [{component}] does '
                                          'not have a valid component id.'
                                          .format(component=component))

        # checking whether is there any registered component with the same id.
        if component.get_id() in self._components.keys():
            replace = options.get('replace', False)

            if replace is not True:
                raise DuplicateComponentIDError('There is another registered component with '
                                                'id [{id}] but "replace" option is not set, so '
                                                'component [{instance}] could not be registered.'
                                                .format(id=component.get_id(),
                                                        instance=component))

            old_instance = self._components[component.get_id()]

            # we should update all list and dict attributes and also those that their name
            # starts with two consecutive underscores of new component with values
            # from old_instance to prevent loss of any attribute value (for example values
            # that has been added using decorators).
            # this has an obvious caveat, and it is that child classes could not do
            # any customizations on these attributes in their `__init__` method.
            component = self._set_component_attributes(old_instance, component)

            print_warning('Component [{old_instance}] is going to be replaced by [{new_instance}].'
                          .format(old_instance=old_instance, new_instance=component))

        self._components[component.get_id()] = component

    def remove_component(self, component_id):
        """
        removes application component with given id.

        :param tuple[str, object] component_id: component id to be removed.

        :raises ComponentAttributeError: component attribute error.
        """

        if component_id not in self._components:
            raise ComponentAttributeError('Component [{component_id}] is not '
                                          'available in application components.'
                                          .format(component_id=component_id))

        self._components.pop(component_id)

    def _get_safe_current_request(self):
        """
        gets current request in a safe manner.

        meaning that if there is no request in current context,
        returns None instead of raising an error.

        :rtype: CoreRequest
        """

        if has_request_context() is True:
            return flask_request

        return None

    def get_component(self, component_name, **options):
        """
        gets the specified application component.

        :param str component_name: component name.

        :keyword object component_custom_key: component custom key.
                                              if not provided, tries to get it from
                                              request object, if not found,
                                              `DEFAULT_COMPONENT_KEY` will be used.

        :raises InvalidComponentNameError: invalid component name error.
        :raises ComponentAttributeError: component attribute error.

        :rtype: Component
        """

        component_custom_key = options.get('component_custom_key', None)

        if component_custom_key is None:
            component_custom_key = self._extract_component_custom_key()

        # checking whether is there any custom implementation.
        component_custom_id = \
            Component.make_component_id(component_name,
                                        component_custom_key=component_custom_key)

        if component_custom_id in self._components.keys():
            return self._components[component_custom_id]

        # getting default component.
        component_default_id = Component.make_component_id(component_name)
        return self._components[component_default_id]

    def _extract_component_custom_key(self):
        """
        gets `component_custom_key` from current request.

        note that if application is in any state lower than `READY`,
        it always returns the default component key.

        :rtype: object
        """

        # before application reaches the `READY` state,
        # we should always use the default component key.
        if self.get_status() in (ApplicationStatusEnum.INITIALIZING,
                                 ApplicationStatusEnum.LOADING):

            return DEFAULT_COMPONENT_KEY

        # this method is the only place that we should access request
        # directly and not from session services, because of circular calling problem.
        client_request = self._get_safe_current_request()
        if client_request is not None:
            return client_request.component_custom_key

        return DEFAULT_COMPONENT_KEY

    def _load(self, **options):
        """
        loads application configs and components.
        """

        self._set_status(ApplicationStatusEnum.LOADING)
        self._resolve_required_paths(**options)
        self._load_environment_variables()
        packaging_services.load_components(**options)
        self._load_version()

        # calling `after_application_loaded` method of all registered hooks.
        self._after_application_loaded()

        # calling `application_initialized` method of all registered hooks.
        self._application_initialized()
        self._set_status(ApplicationStatusEnum.READY)

        if self.is_scripting_mode() is False:
            self._prepare_runtime_data()

    def _resolve_required_paths(self, **options):
        """
        resolves all required paths for application.
        """

        self._resolve_settings_path(**options)
        self._resolve_migrations_path(**options)
        self._resolve_locale_path(**options)

    def _load_version(self):
        """
        loads application version.

        it looks for a `__version__` attribute in main package of application.
        this method could be overridden in subclasses to change behavior.
        """

        package_name = path_utils.get_main_package_name(self.__module__)
        package = packaging_services.load(package_name)
        version = getattr(package, '__version__', None)

        if version is not None:
            self._version = version

    def _load_configs(self, **options):
        """
        loads all configurations related to application package.
        """

        config_services.load_configurations(*self.CONFIG_STORES, **options)
        for store_name in self.CONFIG_STORES:
            config_dict = config_services.get_all(store_name, **options)
            self.configure(config_dict)

    @setupmethod
    def load_configs(self, **options):
        """
        loads all configurations related to application package.

        normally, you should not call this method manually.
        """

        self._load_configs(**options)

    def _configure(self, config_store):
        """
        configures the application with given dict.

        all keys will be converted to uppercase for flask compatibility.

        :param dict config_store: a dictionary containing configuration key/values.
        """

        upper_dict = make_key_upper(config_store)
        self.config.update(upper_dict)

    @setupmethod
    def configure(self, config_store):
        """
        configures the application with given dict.

        all keys will be converted to uppercase for flask compatibility.

        :param dict config_store: a dictionary containing configuration key/values.
        """

        self._configure(config_store)

    def run(self, host=None, port=None, debug=None,
            load_dotenv=True, **options):
        """
        runs the Application instance.

        :param str host: the hostname to listen on. set this to `0.0.0.0` to
                         have the server available externally as well. defaults
                         to `127.0.0.1` or the host in the `SERVER_NAME` config
                         variable if present.

        :param int port: the port of the webserver. defaults to `5000` or the
                         port defined in the `SERVER_NAME` config variable if present.

        :param bool debug: if given, enable or disable debug mode.

        :param bool load_dotenv: load the nearest `.env` and `.flaskenv`
                                 files to set environment variables. will also
                                 change the working directory to the directory
                                 containing the first file found.

        :keyword bool use_reloader: specifies that the server should automatically
                                    restart the python process if modules were changed.

        :keyword bool use_debugger: specifies that the werkzeug debugging
                                    system should be used.

        :keyword bool threaded: specifies that the process should handle
                                each request in a separate thread.

        :raises ApplicationInScriptingModeError: application in scripting mode error.
        """

        if self.is_scripting_mode() is True:
            raise ApplicationInScriptingModeError('Application has been initialized in '
                                                  'scripting mode, so it could not be run.')

        self._before_application_run()
        self._set_status(ApplicationStatusEnum.RUNNING)
        host, port = self._get_communication_configs(host, port)
        super().run(host, port, debug, load_dotenv, **options)

    def _get_communication_configs(self, host, port):
        """
        gets the host and port to use for application.

        :param str host: host name or ip address.
        :param int port: port number.

        :returns: tuple[str host, int port]
        :rtype: tuple[str, int]
        """

        if host in (None, ''):
            host = config_services.get_active('communication', 'server_host')

        if port is None:
            port = config_services.get_active('communication', 'server_port')

        return host, port

    def dispatch_request(self):
        """
        does the request dispatching.

        matches the URL and returns the return value of the view or
        error handlers. this does not have to be a response object.
        in order to convert the return value to a proper response object,
        call `make_response` function.

        :raises AuthenticationFailedError: authentication failed error.
        :raises ViewFunctionRequiredParamsError: view function required params error.
        """

        # we have to override whole `dispatch_request` method to be able to
        # customize it, because of the flask design that everything is embedded
        # inside the `dispatch_request` method.

        client_request = session_services.get_current_request()
        if client_request.routing_exception is not None:
            self.raise_routing_exception(client_request)

        route = client_request.url_rule

        # if we provide automatic options for this URL and the
        # request came with the OPTIONS method, reply automatically.
        if getattr(route, 'provide_automatic_options', False) \
           and client_request.method == HTTPMethodEnum.OPTIONS:
            return self.make_default_options_response()

        # otherwise call the handler for this route.
        return route.handle(client_request.get_inputs())

    def _authenticate(self, client_request):
        """
        authenticates given request.

        :param CoreRequest client_request: request to be authenticated.

        :raises AuthenticationFailedError: authentication failed error.
        """

        authentication_services.authenticate(client_request)

    def get_current_url_adapter(self):
        """
        gets url adapter of current request.

        if request context is not present, this method raises an error.

        :rtype: MapAdapter
        """

        return request_stack.top.url_adapter

    def make_response(self, rv):
        """
        converts the return value from a view function to an instance of `CoreResponse`.

        note that the `rv` value before passing to base method must
        be a `tuple`, `dict`, `str` or `CoreResponse`. otherwise
        it causes an error.

        :param tuple | dict | str | CoreResponse rv: the return value
                                                     from the view function.

        :rtype: CoreResponse
        """

        body, status_code, headers = response_services.unpack_response(rv)
        if not isinstance(body, self.response_class) and not callable(body):
            mimetype = mimetype_services.get_mimetype(body)
            if mimetype != MIMETypeEnum.HTML and (mimetype != MIMETypeEnum.JSON or
                                                  not isinstance(body, str)):
                body, metadata, paginator = self._paginate_result(body)
                if self._force_json_response is True:
                    body = self._prepare_json(body, metadata=metadata, paginator=paginator)
                elif self.default_response_converter is not None:
                    body = self.default_response_converter(body,
                                                           metadata=metadata,
                                                           mimetype=mimetype,
                                                           paginator=paginator)

                if body is None:
                    body = ''

        response = response_services.pack_response(body, status_code, headers)
        result = super().make_response(response)
        result.original_data = body
        return result

    def _paginate_result(self, body, **options):
        """
        paginates response body if required.

        it returns the paginated result and pagination metadata and also the paginator itself.
        if pagination is not done, it returns the same input and None as metadata and paginator.

        :param tuple | dict | str body: the return value from the view function.

        :returns: tuple[list | tuple | dict | str  items, dict metadata, PaginatorBase paginator]
        :rtype: tuple[list | tuple | dict | str, dict, PaginatorBase]
        """

        paginator = session_services.get_request_context('paginator', None)
        if paginator is not None and isinstance(body, list):
            body, metadata = paginator.paginate(body)
            return body, metadata, paginator

        return body, None, None

    def _prepare_json(self, rv, metadata=None, paginator=None):
        """
        prepares the input value to be convertible to json.

        :param object rv: the return value from the view function.

        :param dict metadata: metadata that should be injected into
                              response for pagination.

        :param PaginatorBase paginator: paginator that has been used.

        :rtype: dict
        """

        if rv is None:
            rv = DTO()
        else:
            rv = self._serialize_result(rv, paginator)

        # we could not return a list as response, so we wrap
        # the result in a dict when we want to return a list.
        if isinstance(rv, list):
            result = OrderedDict()
            if metadata is not None:
                result.update(metadata)
            else:
                result.update(count=len(rv))
            result.update(results=rv)
            rv = result

        # we should wrap all single values into a
        # dict before returning it to client.
        if not isinstance(rv, (tuple, dict, self.response_class)):
            rv = DTO(value=rv)

        return rv

    def _serialize_result(self, rv, paginator=None):
        """
        serializes the return value if needed.

        this method could be overridden in subclasses.
        it must return a dict or a list of dicts on success or
        the same exact input if could not convert it.

        :param object rv: the return value from the view function.
        :param PaginatorBase paginator: paginator that has been used.

        :rtype: dict | list[dict]
        """

        result_schema = session_services.get_request_context('result_schema', None)
        if result_schema is not None:
            return result_schema.filter(rv, paginator=paginator)

        return serializer_services.serialize(rv)

    @setupmethod
    def add_url_rule(self, rule, view_func,
                     provide_automatic_options=None, **options):
        """
        connects a url rule. the provided view_func will be registered with the endpoint.

        if there is another rule with the same url and http methods and `replace=True`
        option is provided, it will be replaced. otherwise an error will be raised.

        a note about endpoint. pyrin will handle endpoint generation on its own.
        so there is no endpoint parameter in this method's signature.
        this is required to be able to handle uniqueness of endpoints and managing them.
        despite flask, pyrin will not require you to define view functions with unique names.
        you could define view functions with the same name in different modules. but to
        ensure the uniqueness of endpoints, pyrin will use the fully qualified name
        of function as endpoint. for example: `pyrin.api.services.create_route`.
        so you could figure out endpoint for any view function to use it in places
        like `url_for` method.

        pyrin handles endpoints this way to achieve these two important features:

        1. dismissal of the need for view function name uniqueness.
           when application grows, it's nonsense to be forced to have
           unique view function names at application level.

        2. managing routes with the same url and http methods, and informing
           the developer about them to prevent accidental replacements. and also
           providing a way to replace a route by another route with the same url
           and http methods if that is what developer actually wants.
           when application grows, it becomes a point of failure when you have no
           idea that you've registered a similar route in multiple places and only
           one of them will be get called based on registration order.

        :param str rule: the url rule as string.

        :param function view_func: the function to call when serving a request to the
                                   provided endpoint.

        :param bool provide_automatic_options: controls whether the `OPTIONS` method should be
                                               added automatically.
                                               this can also be controlled by setting the
                                               `view_func.provide_automatic_options = False`
                                               before adding the rule.

        :keyword str | tuple[str] methods: http methods that this rule should handle.
                                           if not provided, defaults to `GET`.

        :keyword PermissionBase | tuple[PermissionBase] permissions: all required permissions
                                                                     for accessing this route.

        :keyword bool authenticated: specifies that this route could not be accessed
                                     if the requester has not been authenticated.
                                     defaults to True if not provided.

        :keyword bool fresh_auth: specifies that this route could not be accessed
                                  if the requester has not a fresh authentication.
                                  fresh authentication means an authentication that
                                  has been done by providing user credentials to
                                  server. defaults to False if not provided.

        :keyword bool replace: specifies that this route must replace any existing
                               route with the same url and http methods or raise
                               an error if not provided. defaults to False.

        :keyword dict defaults: an optional dict with defaults for other rules with the
                                same endpoint. this is a bit tricky but useful if you
                                want to have unique urls.

        :keyword str subdomain: the subdomain rule string for this rule. If not specified the
                                rule only matches for the `default_subdomain` of the map. if
                                the map is not bound to a subdomain this feature is disabled.

        :keyword bool strict_slashes: override the `Map` setting for `strict_slashes` only for
                                      this rule. if not specified the `Map` setting is used.

        :keyword bool merge_slashes: override `Map.merge_slashes` for this rule.

        :keyword bool build_only: set this to True and the rule will never match but will
                                  create a url that can be build. this is useful if you have
                                  resources on a subdomain or folder that are not handled by
                                  the WSGI application (like static data)

        :keyword str | callable redirect_to: if given this must be either a string
                                             or callable. in case of a callable it's
                                             called with the url adapter that
                                             triggered the match and the values
                                             of the url as keyword arguments and has
                                             to return the target for the redirect,
                                             otherwise it has to be a string with
                                             placeholders in rule syntax.

        :keyword bool alias: if enabled this rule serves as an alias for another rule with
                             the same endpoint and arguments.

        :keyword str host: if provided and the url map has host matching enabled this can be
                           used to provide a match rule for the whole host. this also means
                           that the subdomain feature is disabled.

        :keyword bool websocket: if set to True, this rule is only matches for
                                 websocket (`ws://`, `wss://`) requests. by default,
                                 rules will only match for http requests.
                                 defaults to False if not provided.

        :keyword int max_content_length: max content length that this route could handle,
                                         in bytes. if not provided, it will be set to
                                         `restricted_max_content_length` api config key.
                                         note that this value should be lesser than or equal
                                         to `max_content_length` api config key, otherwise
                                         it will cause an error.

        :keyword int status_code: status code to be returned on successful responses.
                                  defaults to corresponding status code of request's
                                  http method if not provided.

        :note status_code: it could be a value from `InformationResponseCodeEnum`
                           or `SuccessfulResponseCodeEnum` or `RedirectionResponseCodeEnum`.

        :keyword bool strict_status: specifies that it should only consider
                                     the status code as processed if it is from
                                     `InformationResponseCodeEnum` or
                                     `SuccessfulResponseCodeEnum` or
                                     `RedirectionResponseCodeEnum` values. otherwise
                                     all codes from `INFORMATION_CODE_MIN`
                                     to `INFORMATION_CODE_MAX` or from
                                     `SUCCESS_CODE_MIN` to `SUCCESS_CODE_MAX`
                                     or from `REDIRECTION_CODE_MIN` to
                                     `REDIRECTION_CODE_MAX` will be considered
                                     as processed. defaults to True if not provided.

        :keyword str | list[str] environments: a list of all environments that this
                                               route must be exposed on them.
                                               the values could be from all available
                                               environments in environments config store.
                                               for example: `production`, `development`.
                                               if not provided, the route will be exposed
                                               on all environments.

        :keyword ResultSchema | type[ResultSchema] result_schema: result schema to be used
                                                                  to filter results. it could
                                                                  be an instance or a type
                                                                  of `ResultSchema` class.

        :keyword bool indexed: specifies that list results must
                               include an extra field as row index.
                               the name of the index field and the initial value
                               of index could be provided by `index_name` and
                               `start_index` respectively. `indexed` keyword has
                               only effect if the returning result contains a list
                               of objects.

        :keyword str index_name: name of the extra field to contain
                                 the row index of each result. if not provided
                                 defaults to `row_num` value.

        :keyword int start_index: the initial value of row index. if not
                                  provided, starts from 1.

        :keyword SECURE_TRUE | SECURE_FALSE readable: specifies that any column or attribute
                                                      which has `allow_read=False` or its name
                                                      starts with underscore `_`, should not
                                                      be included in result dict. defaults to
                                                      `SECURE_TRUE` if not provided. it will
                                                      be used only for entity conversion.
                                                      this value will override the
                                                      corresponding value of `result_schema`
                                                      if provided.

        :keyword int depth: a value indicating the depth for conversion.
                            for example if entity A has a relationship with
                            entity B and there is a list of B in A, if `depth=0`
                            is provided, then just columns of A will be available
                            in result dict, but if `depth=1` is provided, then all
                            B entities in A will also be included in the result dict.
                            actually, `depth` specifies that relationships in an
                            entity should be followed by how much depth.
                            note that, if `columns` is also provided, it is required to
                            specify relationship property names in provided columns.
                            otherwise they won't be included even if `depth` is provided.
                            defaults to `default_depth` value of database config store.
                            please be careful on increasing `depth`, it could fail
                            application if set to higher values. choose it wisely.
                            normally the maximum acceptable `depth` would be 2 or 3.
                            there is a hard limit for max valid `depth` which is set
                            in `ConverterMixin.MAX_DEPTH` class variable. providing higher
                            `depth` value than this limit, will cause an error.
                            it will be used only for entity conversion.
                            this value will override the corresponding value of
                            `result_schema` if provided.

        :keyword bool no_cache: a value indicating that the response returning from this route
                                must have a `Cache-Control: no-cache` header. this header will
                                be automatically added. defaults to False if not provided.

        :keyword int request_limit: number of allowed requests to this
                                    route before it unregisters itself.
                                    defaults to None if not Provided.

        :keyword int lifetime: number of seconds in which this route must remain
                               responsive after initial registration. after this
                               period, the route will unregister itself.
                               defaults to None if not provided.

        :note request_limit, lifetime: if both of these values are provided, this
                                       route will unregister itself if any of
                                       these two conditions are met.

        :keyword bool paged: specifies that this route should return paginated results.
                             defaults to False if not provided.

        :keyword int page_size: default page size for this route.
                                defaults to `default_page_size` from
                                `database` config store if not provided.

        :keyword int max_page_size: maximum page size that client is allowed
                                    to request for this route. defaults to
                                    `max_page_size` from `database` configs store
                                    if not provided.

        :keyword bool cors_enabled: specifies that cross origin resource sharing is enabled.
                                    if not provided, it will be get from cors config store.

        :keyword bool cors_always_send: specifies that cors headers must be included in
                                        response even if the request does not have origin header.
                                        if not provided, it will be get from cors config store.

        :keyword list[str] cors_allowed_origins: a list of extra allowed origins to be used
                                                 in conjunction with default allowed ones.

        :keyword list[str] cors_exposed_headers: extra exposed headers to be combined
                                                 with default ones.

        :keyword list[str] cors_allowed_headers: extra allowed headers to be combined
                                                 with default ones.

        :keyword bool cors_allow_credentials: specifies that browsers are allowed to pass
                                              response headers to front-end javascript code
                                              if the route is authenticated.
                                              if not provided, it will be get from cors config
                                              store.

        :keyword int cors_max_age: maximum number of seconds to cache results.
                                   if not provided, it will be get from cors config store.

        :raises DuplicateRouteURLError: duplicate route url error.
        :raises OverwritingEndpointIsNotAllowedError: overwriting endpoint is not allowed error.
        :raises PageSizeLimitError: page size limit error.
        :raises MaxContentLengthLimitMismatchError: max content length limit mismatch error.
        :raises InvalidViewFunctionTypeError: invalid view function type error.
        :raises InvalidResultSchemaTypeError: invalid result schema type error.
        :raises InvalidResponseStatusCodeError: invalid response status code error.
        """

        environments = options.get('environments')
        if self._should_expose_route(rule, view_func,
                                     environments) is not True:
            return

        endpoint = options.get('endpoint')
        if endpoint in (None, ''):
            endpoint = self.generate_endpoint(view_func, **options)
            options.update(endpoint=endpoint)

        methods = options.pop('methods', None)
        if methods is not None:
            methods = misc_utils.make_iterable(methods, tuple)

        # if the methods are not given and the view_func object knows its
        # methods we can use that instead. if neither exists, we go with
        # a tuple of only `GET` as default.
        if methods is None:
            methods = getattr(view_func, 'methods', None) or (HTTPMethodEnum.GET,)

        methods = set(item.upper() for item in methods)
        required_methods = set(getattr(view_func, 'required_methods', ()))

        # starting with flask 0.8 the view_func object can disable and
        # force-enable the automatic options handling.
        if provide_automatic_options is None:
            provide_automatic_options = getattr(
                view_func, 'provide_automatic_options', None)

        if provide_automatic_options is None:
            if HTTPMethodEnum.OPTIONS not in methods:
                provide_automatic_options = True
                required_methods.add(HTTPMethodEnum.OPTIONS)
            else:
                provide_automatic_options = False

        # add the required methods now.
        methods |= required_methods

        # we have to put `view_function=view_func` into options to be able to deliver it
        # to route initialization in the super method. that's because of flask design
        # that does not forward all params to inner method calls.
        options.update(view_function=view_func)

        route = self.url_rule_class(rule, methods=methods, **options)
        route.provide_automatic_options = provide_automatic_options
        self._add_to_map(route, **options)

        old_func = self.view_functions.get(endpoint)
        if old_func is not None and old_func != view_func:
            old_name = function_utils.get_fully_qualified_name(old_func)
            new_name = function_utils.get_fully_qualified_name(view_func)
            raise OverwritingEndpointIsNotAllowedError('View function [{new_function}] '
                                                       'is overwriting an existing endpoint '
                                                       '[{endpoint}] with registered view '
                                                       'function [{old_function}]. this could '
                                                       'be the result of manually modified '
                                                       'endpoints. pyrin will handle endpoint '
                                                       'generation on its own. so it is '
                                                       'recommended not to define or modify '
                                                       'endpoints manually.'
                                                       .format(new_function=new_name,
                                                               endpoint=endpoint,
                                                               old_function=old_name))

        self.view_functions[endpoint] = view_func

    def _should_expose_route(self, url, view_func, environments):
        """
        gets a value indicating that given route must be exposed on current environment.

        :param str url: the url rule as string.

        :param function view_func: the function to call when serving a request to the
                                   provided endpoint.

        :param str | list[str] environments: a list of all environments that this
                                             route must be exposed on them.
                                             the values could be from all available
                                             environments in environments config store.
                                             for example: `production`, `development`.
                                             if not provided, the route will be exposed
                                             on all environments.

        :rtype: bool
        """

        environments = misc_utils.make_iterable(environments, list)
        should_expose = True
        if len(environments) > 0:
            current_env = config_services.get_active_section_name('environment')
            should_expose = current_env in environments
            if should_expose is not True:
                func_name = function_utils.get_fully_qualified_name(view_func)
                print_warning('Route [{url}] on view function [{func}] will only be '
                              'exposed on {env} environments. the current environment '
                              'is [{current}].'.format(url=url, func=func_name,
                                                       env=environments, current=current_env))
        return should_expose

    def generate_endpoint(self, func, **options):
        """
        generates endpoint for given function.

        pyrin will assume endpoint as function's fully qualified name.
        this method could be overridden in subclasses to change the endpoint generation.

        :param function func: function to generate endpoint for it.

        :rtype: str
        """

        return function_utils.get_fully_qualified_name(func)

    def _add_to_map(self, route, **options):
        """
        adds the given route into map.

        :param RouteBase route: route instance to be added into map.

        :keyword bool replace: specifies that this route must replace
                               any existing route with the same url and http
                               methods or raise an error if not provided.
                               defaults to False.

        :raises DuplicateRouteURLError: duplicate route url error.
        """

        replace = options.get('replace', False)
        existing_routes = self.url_map.get_routes_by_url(route.rule)

        duplicate_methods = DTO()
        for item in existing_routes:
            duplicated = item.get_duplicate_methods(route.methods)
            if len(duplicated) > 0:
                duplicated_for_show = list(duplicated)
                old_func = function_utils.get_fully_qualified_name(item.view_function)
                new_func = function_utils.get_fully_qualified_name(route.view_function)
                if replace is True:
                    duplicate_methods[tuple(duplicated)] = item
                    print_warning('Registered route with url [{url}] for http methods '
                                  '{methods} on view function [{old_func}] is going to be '
                                  'replaced by a new route on view function [{new_func}].'
                                  .format(url=route.rule, methods=duplicated_for_show,
                                          old_func=old_func, new_func=new_func))
                else:
                    raise DuplicateRouteURLError('There is another registered route with the '
                                                 'same url [{url}] and http methods {methods} '
                                                 'on view function [{old_func}], but "replace" '
                                                 'option is not set, so the new route on view '
                                                 'function [{new_func}] could not be registered.'
                                                 .format(url=route.rule,
                                                         methods=duplicated_for_show,
                                                         old_func=old_func, new_func=new_func))

        if len(duplicate_methods) > 0:
            for methods, duplicate_route in duplicate_methods.items():
                duplicate_route.remove_methods(methods)
                if not duplicate_route.is_operational:
                    self.url_map.remove(duplicate_route)

        self.url_map.add(route)

    @setupmethod
    def register_route_factory(self, factory):
        """
        registers a route factory as application url rule class.

        :param callable factory: route factory.
                                 it could be a class or a factory method.

        :raises InvalidRouteFactoryTypeError: invalid route factory type error.
        """

        if not callable(factory):
            raise InvalidRouteFactoryTypeError('Input parameter [{factory}] is not callable.'
                                               .format(factory=factory))

        old_factory = misc_utils.try_get_fully_qualified_name(self.url_rule_class)
        new_factory = misc_utils.try_get_fully_qualified_name(factory)
        print_warning('Registered route factory [{old_factory}] is '
                      'going to be replaced by a new route factory [{new_factory}].'
                      .format(old_factory=old_factory,
                              new_factory=new_factory))

        self.url_rule_class = factory

    def get_current_route_factory(self):
        """
        gets current route factory in use.

        :rtype: callable
        """

        return self.url_rule_class

    def get_default_settings_path(self):
        """
        gets the pyrin default settings path.

        :rtype: str
        """

        result = self.get_context(self.PYRIN_DEFAULT_SETTINGS_CONTEXT_KEY, default=None)
        if result is not None:
            return result

        self._resolve_default_settings_path()
        return self.get_context(self.PYRIN_DEFAULT_SETTINGS_CONTEXT_KEY)

    def get_settings_path(self):
        """
        gets the application settings path.

        :rtype: str
        """

        result = self.get_context(self.SETTINGS_CONTEXT_KEY, default=None)
        if result is not None:
            return result

        self._resolve_settings_path()
        return self.get_context(self.SETTINGS_CONTEXT_KEY)

    def get_migrations_path(self):
        """
        gets the application migrations path.

        :rtype: str
        """

        result = self.get_context(self.MIGRATIONS_CONTEXT_KEY, default=None)
        if result is not None:
            return result

        self._resolve_migrations_path()
        return self.get_context(self.MIGRATIONS_CONTEXT_KEY)

    def get_locale_path(self):
        """
        gets the application locale path.

        :rtype: str
        """

        result = self.get_context(self.LOCALE_CONTEXT_KEY, default=None)
        if result is not None:
            return result

        self._resolve_locale_path()
        return self.get_context(self.LOCALE_CONTEXT_KEY)

    def get_application_main_package_path(self):
        """
        gets the application main package path.

        :rtype: str
        """

        result = self.get_context(self.APPLICATION_PATH_CONTEXT_KEY, default=None)
        if result is not None:
            return result

        self._resolve_application_main_package_path()
        return self.get_context(self.APPLICATION_PATH_CONTEXT_KEY)

    def get_pyrin_root_path(self):
        """
        gets pyrin root path in which pyrin package is located.

        :rtype: str
        """

        result = self.get_context(self.ROOT_PYRIN_PATH_CONTEXT_KEY, default=None)
        if result is not None:
            return result

        self._resolve_pyrin_root_path()
        return self.get_context(self.ROOT_PYRIN_PATH_CONTEXT_KEY)

    def get_application_root_path(self):
        """
        gets the application root path in which application package is located.

        :rtype: str
        """

        result = self.get_context(self.ROOT_APPLICATION_PATH_CONTEXT_KEY, default=None)
        if result is not None:
            return result

        self._resolve_application_root_path()
        return self.get_context(self.ROOT_APPLICATION_PATH_CONTEXT_KEY)

    def get_pyrin_main_package_path(self):
        """
        gets pyrin main package path.

        :rtype: str
        """

        result = self.get_context(self.PYRIN_PATH_CONTEXT_KEY, default=None)
        if result is not None:
            return result

        self._resolve_pyrin_main_package_path()
        return self.get_context(self.PYRIN_PATH_CONTEXT_KEY)

    def get_working_directory(self):
        """
        gets working directory path of application.

        working directory is where the root application and test package are resided.
        this is required when application starts from any of test applications.
        then we should move root path up, to the correct root to be able to
        include real application packages too.
        if the application has been started from real application, this method
        returns the same result as `get_application_root_path()` method.

        :rtype: str
        """

        return packaging_services.get_working_directory(self.get_application_root_path())

    def get_configs(self):
        """
        gets a shallow copy of application's configuration dictionary.

        :rtype: dict
        """

        return self.config.copy()

    def _resolve_settings_path(self, **options):
        """
        resolves the application settings path.

        the resolved path will be accessible by `SETTINGS_CONTEXT_KEY`
        inside application context.

        :keyword str settings_directory: settings directory name.
                                         if not provided, defaults to `settings`.
        """

        directory = options.get('settings_directory', 'settings')
        directory = os.path.split(directory)
        main_package_path = self.get_application_main_package_path()
        settings_path = os.path.join(main_package_path, *directory)
        settings_path = os.path.abspath(settings_path)

        if not os.path.isdir(settings_path):
            print_warning('Application settings path [{path}] does not exist. '
                          'pyrin default settings will be used. you could change '
                          'any of default setting files inside [{path}] on your preference. '
                          'DO NOT use pyrin default settings in production!'
                          .format(path=settings_path))

        self.add_context(self.SETTINGS_CONTEXT_KEY, settings_path)

    def _resolve_default_settings_path(self):
        """
        resolves the pyrin default settings path.

        the resolved path will be accessible by `PYRIN_DEFAULT_SETTINGS_CONTEXT_KEY`
        inside application context.
        """

        pyrin_main_package_path = self.get_pyrin_main_package_path()
        settings_path = os.path.join(pyrin_main_package_path, 'settings', 'default')
        settings_path = os.path.abspath(settings_path)
        self.add_context(self.PYRIN_DEFAULT_SETTINGS_CONTEXT_KEY, settings_path)

    def _resolve_migrations_path(self, **options):
        """
        resolves the application migrations path.

        the resolved path will be accessible by `MIGRATIONS_CONTEXT_KEY`
        inside application context.

        :keyword str migrations_directory: migrations directory name.
                                           if not provided, defaults to `migrations`.
        """

        directory = options.get('migrations_directory', 'migrations')
        directory = os.path.split(directory)
        main_package_path = self.get_application_main_package_path()
        migrations_path = os.path.join(main_package_path, *directory)
        migrations_path = os.path.abspath(migrations_path)

        self.add_context(self.MIGRATIONS_CONTEXT_KEY, migrations_path)

    def _resolve_locale_path(self, **options):
        """
        resolves the application locale path.

        the resolved path will be accessible by `LOCALE_CONTEXT_KEY`
        inside application context.

        :keyword str locale_directory: locale directory name.
                                       if not provided, defaults to `locale`.
        """

        directory = options.get('locale_directory', 'locale')
        directory = os.path.split(directory)
        main_package_path = self.get_application_main_package_path()
        locale_path = os.path.join(main_package_path, *directory)
        locale_path = os.path.abspath(locale_path)

        self.add_context(self.LOCALE_CONTEXT_KEY, locale_path)

    def _resolve_application_main_package_path(self, **options):
        """
        resolves the application main package path.

        it registers it in application context with `APPLICATION_PATH_CONTEXT_KEY` key.
        """

        main_package_path = None
        if self._import_name is None:
            main_package_path = path_utils.get_main_package_path(self.__module__)
        else:
            main_package_path = path_utils.get_package_path(self._import_name)

        self.add_context(self.APPLICATION_PATH_CONTEXT_KEY, main_package_path)

    def _resolve_pyrin_main_package_path(self, **options):
        """
        resolves pyrin main package path.

        it registers it in application context with `PYRIN_PATH_CONTEXT_KEY` key.
        """

        pyrin_main_package = path_utils.get_pyrin_main_package_path()
        self.add_context(self.PYRIN_PATH_CONTEXT_KEY, pyrin_main_package)

    def _resolve_pyrin_root_path(self, **options):
        """
        resolves pyrin root path.

        it registers it in application context with `ROOT_PYRIN_PATH_CONTEXT_KEY` key.
        """

        main_package_path = self.get_pyrin_main_package_path()
        root_path = os.path.join(main_package_path, '..')
        root_path = os.path.abspath(root_path)
        self.add_context(self.ROOT_PYRIN_PATH_CONTEXT_KEY, root_path)

    def _resolve_application_root_path(self, **options):
        """
        resolves application root path.

        it registers it in application context with `ROOT_APPLICATION_PATH_CONTEXT_KEY` key.
        """

        main_package_path = self.get_application_main_package_path()
        root_path = os.path.join(main_package_path, '..')
        root_path = os.path.abspath(root_path)
        self.add_context(self.ROOT_APPLICATION_PATH_CONTEXT_KEY, root_path)

    def _load_environment_variables(self):
        """
        loads all environment variables defined in a `.env` file.

        it must be in application root path. if the
        file does not exist, it will be ignored.
        """

        root_path = self.get_application_root_path()
        env_file = os.path.join(root_path, '.env')

        if not os.path.isfile(env_file):
            print_warning('Could not find ".env" file in application root path [{root_path}].'
                          .format(root_path=root_path))
            return

        dotenv.load_dotenv(env_file)

    def process_response(self, response):
        """
        this method is overridden to add required attributes into response object.

        :param CoreResponse response: response object.

        :rtype: CoreResponse
        """

        client_request = session_services.get_current_request()
        response.request_date = client_request.request_date
        response.request_id = client_request.request_id
        response.user = session_services.get_current_user()

        return super().process_response(response)

    def full_dispatch_request(self):
        """
        dispatches the request and on top of that performs request pre and postprocessing.

        as well as http exception catching and error handling.
        this method has been overridden to log before and after request dispatching.
        """

        response = None
        client_request = session_services.get_current_request()
        process_start_time = time()
        logging_services.info('Request received with params: [{params}] '
                              'and headers: [{headers}].',
                              interpolation_data=
                              dict(params=self._get_request_data_for_logging(client_request),
                                   headers=client_request.headers))
        try:
            self._validate_request(client_request)
        except Exception as error:
            logging_services.exception(str(error))
            response = response_services.make_exception_response(error)
            response = self.make_response(response)

        if response is None:
            try:
                self._authenticate(client_request)
            except Exception as error:
                logging_services.exception(str(error))

            response = super().full_dispatch_request()

        response = self._finalize_transaction(response)

        process_end_time = time()
        logging_services.info('Request executed in [{time} ms].'
                              .format(time='{:0.3f}'
                                      .format((process_end_time - process_start_time) * 1000)))

        logging_services.debug('Response [{response}] returned with result: [{result}] '
                               'and headers: [{headers}].',
                               interpolation_data=
                               dict(response=response,
                                    result=self._get_response_data_for_logging(response),
                                    headers=response.headers))

        return response

    def finalize_request(self, rv, from_error_handler=False):
        """
        given the return value from a view function this finalizes the request.

        by converting it into a response and invoking the postprocessing functions.
        this is invoked for both normal request dispatching as well as error handlers.

        because this means that it might be called as a result of a failure a special
        safe mode is available which can be enabled with the `from_error_handler` flag.
        if enabled, failures in response processing will be logged and otherwise ignored.

        this method is overridden to inject custom response headers using application hook.

        :param object | tuple | CoreResponse rv: response from view function or error handlers.

        :param bool from_error_handler: specifies that this method is called as a result
                                        of a failure. defaults to False if not provided.

        :rtype: CoreResponse
        """

        body, status_code, headers = response_services.unpack_response(rv)

        extra_headers = self.headers_class()
        client_request = session_services.get_current_request()
        self._provide_response_headers(extra_headers, client_request.endpoint,
                                       status_code, client_request.method,
                                       url=client_request.path,
                                       user=client_request.user)
        extra_headers.extend(headers or {})
        response = response_services.pack_response(body, status_code, extra_headers)

        return super().finalize_request(response, from_error_handler=from_error_handler)

    def _get_request_data_for_logging(self, request):
        """
        gets the request data for logging.

        if request content length is larger than allowed value, it will be ignored.

        :param CoreRequest request: request instance.

        :rtype: dict
        """

        max_length = config_services.get_active('logging', 'max_request_size')
        if request.safe_content_length > max_length:
            return 'Request payload [{size} bytes] is too large for logging.' \
                .format(size=request.safe_content_length)

        return request.get_inputs(silent=True)

    def _get_response_data_for_logging(self, response):
        """
        gets the response data for logging.

        if response content length is larger than allowed value, it will be ignored.

        :param CoreResponse response: response instance.

        :rtype: dict | object
        """

        max_length = config_services.get_active('logging', 'max_response_size')
        if response.safe_content_length > max_length:
            return 'Response payload [{size} bytes] is too large for logging.' \
                .format(size=response.safe_content_length)

        log_all_types = config_services.get_active('logging', 'log_all_response_types')
        if response.content_type != MIMETypeEnum.JSON and \
                log_all_types is not True:
            return 'Response payload type [{content_type}] is ignored for logging.' \
                .format(content_type=response.content_type)

        return response.original_data

    def _set_component_attributes(self, old_instance, new_instance):
        """
        replaces required component attributes from old instance into new instance.

        it will replace all list and dict attributes from old instance into new instance.
        all the attributes which their name starts with two underscores (private attributes)
        will also be replaced from old instance into new instance.

        :param Component old_instance: old component instance to get attributes from.
        :param Component new_instance: new component instance to set its attributes.

        :rtype: Component
        """

        if old_instance is None or new_instance is None:
            return new_instance

        all_attributes = vars(old_instance)
        required_attributes = DTO()
        for attribute_name in all_attributes:
            if self._should_fill_from_parent(attribute_name,
                                             all_attributes[attribute_name]) is True:
                required_attributes[attribute_name] = all_attributes[attribute_name]

        return misc_utils.set_attributes(new_instance, **required_attributes)

    def _should_fill_from_parent(self, name, value):
        """
        gets a value indicating that given attribute value must be filled from parent.

        it is used on initializing subclassed managers (components) in extended packages.
        all list and dict values will result in True.
        all the attributes which their name starts with two underscores (private attributes)
        will also result in True.
        other attributes result in False.

        :param str name: attribute name to be checked.
        :param object value: attribute value to be checked.

        :rtype: bool
        """

        if '__' in name:
            return True

        return isinstance(value, (list, dict))

    def _after_application_loaded(self):
        """
        this method will call `after_application_loaded` method of all registered hooks.
        """

        for hook in self._get_hooks():
            hook.after_application_loaded()

    def _application_initialized(self):
        """
        this method will call `application_initialized` method of all registered hooks.
        """

        for hook in self._get_hooks():
            hook.application_initialized()

    def get_application_name(self):
        """
        gets the application name.

        it is actually the application main package name.
        first, it gets the import name of application if available.
        otherwise gets the first part of application package name

        :rtype: str
        """

        return self._import_name or path_utils.get_main_package_name(self.__module__)

    def _prepare_termination(self, signal_number):
        """
        prepares for termination.

        :param int signal_number: signal number that caused termination.
        """

        self._set_status(ApplicationStatusEnum.TERMINATED)

    def _application_status_changed(self, old_status, new_status):
        """
        this method will call `application_status_changed` method of all registered hooks.

        :param str old_status: old application status.
        :param str new_status: new application status.

        :enum status:
            INITIALIZING = 'Initializing'
            LOADING = 'Loading'
            READY = 'Ready'
            RUNNING = 'Running'
            TERMINATED = 'Terminated'
        """

        for hook in self._get_hooks():
            hook.application_status_changed(old_status, new_status)

    def _before_application_run(self):
        """
        this method will call `before_application_run` method of all registered hooks.

        note that this method will not get called when
        application starts in scripting mode.
        """

        for hook in self._get_hooks():
            hook.before_application_run()

    def _prepare_runtime_data(self):
        """
        this method will call `prepare_runtime_data` method of all registered hooks.

        this method will commit any changes of each hook to database, so you should
        not commit anything in the hooks. if you do commit manually, unexpected
        behaviors may occur.

        note that this method will not get called when application starts in scripting mode.
        """

        for hook in self._get_hooks():
            with atomic_context(expire_on_commit=True):
                hook.prepare_runtime_data()

    def _provide_response_headers(self, headers, endpoint,
                                  status_code, method, **options):
        """
        this method will call `provide_response_headers` method of all registered hooks.

        :param dict | Headers headers: current response headers.

        :param str endpoint: the endpoint of the route that
                             handled the current request.
                             by default, it is the fully qualified
                             name of the view function.

        :param int status_code: response status code.
                                it could be None if not provided.

        :param str method: the http method of current request.

        :keyword str url: the url of the route that handled this request.

        :keyword user: the user of current request.
                       it could be None.
        """

        for hook in self._get_hooks():
            hook.provide_response_headers(headers, endpoint,
                                          status_code, method, **options)

    def _finalize_transaction(self, response, **options):
        """
        this method will call `finalize_transaction` of all registered hooks.

        :param CoreResponse response: response object.

        :rtype: CoreResponse
        """

        result_response = response
        for hook in self._get_hooks():
            try:
                result = hook.finalize_transaction(result_response, **options)
                if result is not None:
                    if not isinstance(result, self.response_class):
                        result = self.make_response(result)
                    result_response = result
            except Exception as error:
                logging_services.exception(str(error))

        return result_response

    def _validate_request(self, request, **options):
        """
        this method will call `validate_request` method of all registered hooks.

        :param CoreRequest request: current request instance.
        """

        for hook in self._get_hooks():
            hook.validate_request(request, **options)

    def get_application_version(self):
        """
        gets application version.

        :rtype: str
        """

        return self._version

    def get_pyrin_version(self):
        """
        gets pyrin version.

        :rtype: str
        """

        return pyrin.__version__



# -*- coding: utf-8 -*-
"""
deserializer handlers pool module.
"""

import re

from sqlalchemy.pool import Pool, NullPool, AssertionPool, QueuePool, \
    SingletonThreadPool, StaticPool

from pyrin.converters.deserializer.decorators import deserializer
from pyrin.converters.deserializer.handlers.base import StringPatternDeserializerBase


@deserializer()
class PoolDeserializer(StringPatternDeserializerBase):
    """
    pool deserializer class.
    """

    # matches the pool class inside string.
    # example: Pool, NullPool, AssertionPool, QueuePool, SingletonThreadPool, StaticPool
    # matching is case-insensitive.
    POOL_REGEX = re.compile(r'^Pool$', re.IGNORECASE)
    NULL_POOL_REGEX = re.compile(r'^NullPool$', re.IGNORECASE)
    ASSERTION_POOL_REGEX = re.compile(r'^AssertionPool$', re.IGNORECASE)
    QUEUE_POOL_REGEX = re.compile(r'^QueuePool$', re.IGNORECASE)
    SINGLETON_THREAD_POOL_REGEX = re.compile(r'^SingletonThreadPool$', re.IGNORECASE)
    STATIC_POOL_REGEX = re.compile(r'^StaticPool$', re.IGNORECASE)

    def __init__(self, **options):
        """
        creates an instance of PoolDeserializer.

        :keyword list[tuple[Pattern, int, int]] accepted_formats: a list of custom accepted
                                                                  formats and their min and
                                                                  max length for pool
                                                                  deserialization.

        :note accepted_formats: list[tuple[Pattern format, int min_length, int max_length]]
        """

        options.update(internal=True)
        super().__init__(**options)

        self._converter_map = self._get_converter_map()

    def _deserialize(self, value, **options):
        """
        deserializes the given value.

        :param str value: value to be deserialized.

        :keyword Pattern matching_pattern: the pattern that has matched the value.

        :rtype: type[Pool]
        """

        pattern = options.pop('matching_pattern')
        return self._converter_map[pattern]

    @property
    def default_formats(self):
        """
        gets default accepted formats that this deserializer could deserialize value from.

        :returns: list[tuple[Pattern format, int min_length, int max_length]]
        :rtype: list[tuple[Pattern, int, int]]
        """

        return [(self.POOL_REGEX, 4, 4),
                (self.NULL_POOL_REGEX, 8, 8),
                (self.ASSERTION_POOL_REGEX, 13, 13),
                (self.QUEUE_POOL_REGEX, 9, 9),
                (self.SINGLETON_THREAD_POOL_REGEX, 19, 19),
                (self.STATIC_POOL_REGEX, 10, 10)]

    def _get_converter_map(self):
        """
        gets converter map dictionary.

        :returns: dict[Pattern format: type pool]
        :rtype: dict
        """

        return {self.POOL_REGEX: Pool,
                self.NULL_POOL_REGEX: NullPool,
                self.ASSERTION_POOL_REGEX: AssertionPool,
                self.QUEUE_POOL_REGEX: QueuePool,
                self.SINGLETON_THREAD_POOL_REGEX: SingletonThreadPool,
                self.STATIC_POOL_REGEX: StaticPool}



# -*- coding: utf-8 -*-
"""
utils unique_id module.
"""

import re
import uuid


# matches the uuid inside string.
# example: 2eaf043b-647a-45fb-b2a4-1d365a8eb548, 74314da0-6e34-11eb-8ce9-000000000000
# it matches all v1, v3, v4 and v5 uuids.
# matching is case-insensitive.
# the valid length is always 36.
UUID_REGEX = re.compile(r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$',
                        re.IGNORECASE)


def generate(**options):
    """
    generates a unique id based on input params.

    :keyword callable generator: uuid generator function.
                                 if not provided, defaults to uuid4.

    :rtype: uuid.UUID
    """

    generator = options.get('generator', uuid.uuid4)
    return generator()


def generate_uuid4():
    """
    generates a unique id using uuid4.

    :rtype: uuid.UUID
    """

    return generate(generator=uuid.uuid4)


def generate_uuid1():
    """
    generates a unique id using uuid1.

    :rtype: uuid.UUID
    """

    return generate(generator=uuid.uuid1)


def get_uuid(value):
    """
    gets the uuid object of given value.

    it raises an error if value is not a valid uuid string.

    :param str value: value to get uuid instance from it.

    :raises ValueError: value error.

    :rtype: uuid.UUID
    """

    return uuid.UUID(value)


def try_get_uuid(value):
    """
    gets the uuid object of given value.

    it returns None if value is not a valid uuid string.

    :param str value: value to get uuid instance from it.

    :rtype: uuid.UUID
    """

    try:
        return get_uuid(value)
    except Exception:
        return None


def try_get_uuid_or_value(value):
    """
    gets the uuid object of given value.

    it returns the same input if value is not a valid uuid string.

    :param str value: value to get uuid instance from it.

    :rtype: uuid.UUID | str
    """

    result = try_get_uuid(value)
    if result is None:
        return value

    return result



# -*- coding: utf-8 -*-
"""
utilities string package.
"""

from pyrin.packaging.base import Package


class UtilitiesStringPackage(Package):
    """
    utilities string package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
security cli module.
"""

import pyrin.security.token.services as token_services
import pyrin.security.hashing.services as hashing_services
import pyrin.security.encryption.services as encryption_services
import pyrin.security.utils.services as security_utils_services

from pyrin.cli.decorators import cli_invoke, cli_group
from pyrin.core.structs import CLI


@cli_group('security')
class SecurityCLI(CLI):
    """
    security cli class.

    this class exposes all security cli commands.
    """

    @cli_invoke
    def token(self, payload=None, **options):
        """
        generates a pair of access and refresh tokens using specified handler.

        the generated token is in the form of `header_hash.payload_hash.signature_hash`
        and each part is encoded using a signing key.

        :keyword dict payload: a dictionary containing key/values as payload.
                               note that for better performance, keep payload
                               as small as possible.

        :keyword str handler_name: name of token handler to be used.
                                   if not provided, default handler
                                   from relevant configs will be used.

        :keyword dict custom_headers: a dictionary containing custom headers.

        :keyword bool is_fresh: indicates that this token is fresh.
                                being fresh means that token is generated by
                                providing user credentials to server.
                                if not provided, defaults to False.

        :raises TokenHandlerNotFoundError: token handler not found error.

        :returns: tuple[str access_token, str refresh_token]
        :rtype: tuple[str, str]
        """

        access = token_services.generate_access_token(payload, **options)
        refresh = token_services.generate_refresh_token(payload, **options)
        return ['ACCESS-TOKEN:', (access,), 'REFRESH-TOKEN:', (refresh,)]

    @cli_invoke
    def hash(self, text, **options):
        """
        gets the hash of input text using a random or specified salt.

        :param str text: text to be hashed.

        :keyword str handler_name: handler name to be used for hash generation.
                                   if not provided, default handler from
                                   relevant configs will be used.

        :keyword bytes salt: salt to be used for hashing.
                             if not provided, a random salt will be generated
                             considering `salt_length` option.

        :keyword str internal_algorithm: internal algorithm to be used
                                         for hashing. if not provided,
                                         default value from relevant
                                         config will be used.

        :keyword int rounds: rounds to perform for generating hash.
                             if not provided, default value from
                             relevant config will be used.

        :keyword int salt_length: salt length to be used for hashing.
                                  if `salt` option is provided, then
                                  this value will be ignored.
                                  if not provided, default value from
                                  relevant config will be used.

        :keyword str prefix: prefix to be used for bcrypt hashing.

        :rtype: str
        """

        hashed_value = hashing_services.generate_hash(text, **options)
        return ['HASH:', (hashed_value,)]

    @cli_invoke
    def encrypt(self, text, **options):
        """
        encrypts the given value using specified handler and returns the encrypted result.

        :param str text: text to be encrypted.

        :keyword str handler_name: handler name to be used for encryption.
                                   if not provided, default handler from
                                   relevant configs will be used.

        :raises EncryptionHandlerNotFoundError: encryption handler not found error.

        :rtype: str
        """

        encrypted = encryption_services.encrypt(text, **options)
        return ['ENCRYPTED:', (encrypted,)]

    @cli_invoke
    def decrypt(self, encrypted, **options):
        """
        decrypts the given full encrypted value using specified
        handler and returns the decrypted result.

        :param str encrypted: full encrypted value to be decrypted.

        :raises InvalidEncryptionValueError: invalid encryption value error.
        :raises EncryptionHandlerNotFoundError: encryption handler not found error.
        :raises DecryptionError: decryption error.

        :rtype: str
        """

        decrypted = encryption_services.decrypt(encrypted, **options)
        return ['DECRYPTED:', (decrypted,)]

    @cli_invoke
    def rsa_key(self, **options):
        """
        generates a pair of public/private rsa keys.

        :keyword int length: key length in bits.
                             if not provided, `rsa_default_key_length`
                             config will be used.

        :returns: tuple[str public_key, str private_key]
        :rtype: tuple[str, str]
        """

        public, private = security_utils_services.generate_rsa_key(**options)
        return ['PUBLIC-KEY:', (public,), 'PRIVATE-KEY:', (private,)]

    @cli_invoke
    def bytes(self, **options):
        """
        gets a secure random bytes with given length.

        the result value should not be decoded to string, because
        it's not a safe-string and may cause an error.
        if you want string representation, use `get_hex` or `get_url_safe` methods.

        :keyword int length: length of random bytes to be get.
                             if not provided, `default_secure_random_size`
                             config will be used.

        :rtype: bytes
        """

        bytes_value = security_utils_services.get_bytes(**options)
        return ['BYTES:', (bytes_value,)]

    @cli_invoke
    def hex(self, **options):
        """
        gets a secure random hex string with given length.

        :keyword int length: length of random string to be get in bytes.
                             if not provided, `default_secure_random_size`
                             config will be used.

        :rtype: str
        """

        hex_value = security_utils_services.get_hex(**options)
        return ['HEX:', (hex_value,)]

    @cli_invoke
    def url_safe(self, **options):
        """
        gets a secure random url-safe string with given length.

        :keyword int length: length of random string to be get in bytes.
                             if not provided, `default_secure_random_size`
                             config will be used.

        :rtype: str
        """

        url_safe = security_utils_services.get_url_safe(**options)
        return ['URL-SAFE:', (url_safe,)]



# -*- coding: utf-8 -*-
"""
celery manager module.
"""

from celery import Celery

import pyrin.configuration.services as config_services
import pyrin.application.services as application_services

from pyrin.core.structs import Manager
from pyrin.task_queues.celery import CeleryPackage
from pyrin.task_queues.celery.structs import ExtendedTask


class CeleryManager(Manager):
    """
    celery manager class.
    """

    package_class = CeleryPackage
    task_class = ExtendedTask

    def __init__(self):
        """
        initializes an instance of CeleryManager.
        """

        super().__init__()

        self._app = self._configure()

    def _configure(self):
        """
        configures celery.
        """

        app = Celery(application_services.get_application_name(),
                     task_cls=self.task_class,
                     set_as_current=True)

        configs = config_services.get_active_section('celery')
        configs.update(worker_hijack_root_logger=False)
        app.config_from_object(configs)
        app.set_default()

        return app

    def get_current_app(self):
        """
        gets current celery application.

        :rtype: Celery
        """

        return self._app



# -*- coding: utf-8 -*-
"""
sentry interface module.
"""

from abc import abstractmethod
from threading import Lock

from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.core.structs import CoreObject, MultiSingletonMeta


class SentryIntegrationSingletonMeta(MultiSingletonMeta):
    """
    sentry integration singleton meta class.

    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class AbstractSentryIntegration(CoreObject, metaclass=SentryIntegrationSingletonMeta):
    """
    abstract sentry integration class.

    all application sentry integrations must be subclassed from this.
    """

    @abstractmethod
    def configure(self, integrations, **options):
        """
        configures this integration.

        :param list[Integration] integrations: list of enabled integrations.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def is_enabled(self):
        """
        gets a value indicating that this integration is enabled in sentry config store.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
caching package.
"""

from pyrin.packaging.base import Package


class CachingPackage(Package):
    """
    caching package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'caching.component'
    CONFIG_STORE_NAMES = ['caching']
    DEPENDS = ['pyrin.configuration',
               'pyrin.database.model']
